import numpy as np

a = 1.0
b = 3.5
order = 81
elem1 = 'C'
elem2 = 'C'

label_2 = '8C-g81-t24-s0'
label_3 = '8C-g81-t24-s72'
coefs_2 = np.load(label_2 + '.npy')
coefs_3 = np.load(label_3 + '.npy')

f = open('C8.txt', 'w')
header_comment = '''# #2bodyarray #3bodyarray
# elem1 elem2 a b order
'''
f.write(header_comment)
header = '''
{twobodyarray} {threebodyarray}
{elem1} {elem2} {a} {b} {order}
'''.format(twobodyarray=1, threebodyarray=1, elem1=elem1, elem2=elem2, a=a, b=b, order=order)
f.write(header)

for c, coef in enumerate(coefs_2):
    f.write('{:.10e} '.format(coef))
    if c % 5 == 4:
        f.write('\n')

f.write('\n')

a = [1.0, 1.0, 0.0]
b = [3.5, 3.5, np.pi]
order = [81, 81, 81]
elem1 = 'C'
elem2 = 'C'
elem3 = 'C'

header_3 = '''
{elem1} {elem2} {elem3} {a1} {a2} {a3} {b1} {b2} {b3:.10e} {order1} {order2} {order3}
'''.format(elem1=elem1, elem2=elem2, elem3=elem3,
        a1=a[0], a2=a[1], a3=a[2], 
        b1=b[0], b2=b[1], b3=b[2], 
        order1=order[0], order2=order[1], order3=order[2])
f.write(header_3)

n = 0
for i in range(coefs_3.shape[0]):
    for j in range(coefs_3.shape[1]):
        for k in range(coefs_3.shape[2]):
            coef = coefs_3[i, j, k]
            f.write('{:.10e} '.format(coef))
            if n % 5 == 4:
                f.write('\n')
            n += 1

f.close()






"""Gaussian process model of the Born Oppenheimer potential energy surface."""
import math
from typing import List, Callable
import numpy as np
from scipy.linalg import solve_triangular
from scipy.optimize import minimize
from flare.env import AtomicEnvironment
from flare.struc import Structure
from flare.gp_algebra import get_ky_and_hyp, get_like_grad_from_mats, \
    get_neg_likelihood, get_neg_like_grad, get_ky_and_hyp_par
from flare.kernels import str_to_kernel
from flare.mc_simple import str_to_mc_kernel


class GaussianProcess:
    """ Gaussian Process Regression Model.

    Implementation is based on Algorithm 2.1 (pg. 19) of
    "Gaussian Processes for Machine Learning" by Rasmussen and Williams"""

    def __init__(self, kernel: Callable,
                 kernel_grad: Callable, hyps: np.ndarray,
                 cutoffs: np.ndarray,
                 hyp_labels: List = None,
                 energy_force_kernel: Callable = None,
                 energy_kernel: Callable = None,
                 opt_algorithm: str = 'L-BFGS-B',
                 maxiter=10, par=False,
                 output=None):
        """Initialize GP parameters and training data."""

        self.kernel = kernel
        self.kernel_grad = kernel_grad
        self.energy_kernel = energy_kernel
        self.energy_force_kernel = energy_force_kernel
        self.kernel_name = kernel.__name__
        self.hyps = hyps
        self.hyp_labels = hyp_labels
        self.cutoffs = cutoffs
        self.algo = opt_algorithm

        self.training_data = []
        self.training_labels = []
        self.training_labels_np = np.empty(0, )
        self.maxiter = maxiter
        self.par = par
        self.output = output

        # Parameters set during training
        self.ky_mat = None
        self.l_mat = None
        self.alpha = None
        self.ky_mat_inv = None
        self.l_mat_inv = None
        self.likelihood = None
        self.likelihood_gradient = None

    # TODO unit test custom range
    def update_db(self, struc: Structure, forces: list,
                  custom_range: List[int] = ()):
        """Given structure and forces, add to training set.

        :param struc: structure to add to db
        :type struc: Structure
        :param forces: list of corresponding forces to add to db
        :type forces: list<float>
        :param custom_range: Indices to use in lieu of the whole structure
        :type custom_range: List[int]
        """

        # By default, use all atoms in the structure
        noa = len(struc.positions)
        update_indices = custom_range or list(range(noa))

        for atom in update_indices:
            env_curr = AtomicEnvironment(struc, atom, self.cutoffs)
            forces_curr = np.array(forces[atom])

            self.training_data.append(env_curr)
            self.training_labels.append(forces_curr)

        # create numpy array of training labels
        self.training_labels_np = self.force_list_to_np(self.training_labels)

    def add_one_env(self, env: AtomicEnvironment,
                    force: np.array, train: bool = False, **kwargs):
        """
        Tool to add a single environment / force pair into the training set
        :param force:
        :param env:
        :param force: (x,y,z) component associated with environment
        :param train:
        :return:
        """
        self.training_data.append(env)
        self.training_labels.append(force)
        self.training_labels_np = self.force_list_to_np(self.training_labels)

        if train:
            self.train(**kwargs)

    @staticmethod
    def force_list_to_np(forces: list) -> np.ndarray:
        """ Convert list of forces to numpy array of forces.

        :param forces: list of forces to convert
        :type forces: list<float>
        :return: numpy array forces
        :rtype: np.ndarray
        """
        forces_np = []

        for force in forces:
            for force_comp in force:
                forces_np.append(force_comp)

        forces_np = np.array(forces_np)

        return forces_np

    def train(self, output=None, custom_bounds=None,
              grad_tol: float = 1e-4,
              x_tol: float = 1e-5,
              line_steps: int = 20):
        """Train Gaussian Process model on training data. Tunes the \
hyperparameters to maximize the likelihood, then computes L and alpha \
(related to the covariance matrix of the training set)."""

        x_0 = self.hyps

        args = (self.training_data, self.training_labels_np,
                self.kernel_grad, self.cutoffs, output,
                self.par)
        res = None

        if self.algo == 'L-BFGS-B':

            # bound signal noise below to avoid overfitting
            bounds = np.array([(1e-6, np.inf)] * len(x_0))
            # bounds = np.array([(1e-6, np.inf)] * len(x_0))
            # bounds[-1] = [1e-6,np.inf]
            # Catch linear algebra errors and switch to BFGS if necessary
            try:
                res = minimize(get_neg_like_grad, x_0, args,
                               method='L-BFGS-B', jac=True, bounds=bounds,
                               options={'disp': False, 'gtol': grad_tol,
                                        'maxls': line_steps,
                                        'maxiter': self.maxiter})
            except:
                print("Warning! Algorithm for L-BFGS-B failed. Changing to "
                      "BFGS for remainder of run.")
                self.algo = 'BFGS'

        if custom_bounds is not None:
            res = minimize(get_neg_like_grad, x_0, args,
                           method='L-BFGS-B', jac=True, bounds=custom_bounds,
                           options={'disp': False, 'gtol': grad_tol,
                                    'maxls': line_steps,
                                    'maxiter': self.maxiter})

        elif self.algo == 'BFGS':
            res = minimize(get_neg_like_grad, x_0, args,
                           method='BFGS', jac=True,
                           options={'disp': False, 'gtol': grad_tol,
                                    'maxiter': self.maxiter})

        elif self.algo == 'nelder-mead':
            res = minimize(get_neg_likelihood, x_0, args,
                           method='nelder-mead',
                           options={'disp': False,
                                    'maxiter': self.maxiter,
                                    'xtol': x_tol})
        if res is None:
            raise RuntimeError("Optimization failed for some reason.")
        self.hyps = res.x
        self.set_L_alpha()
        self.likelihood = -res.fun
        self.likelihood_gradient = -res.jac

    def predict(self, x_t: AtomicEnvironment, d: int) -> [float, float]:
        """Predict force component of an atomic environment and its \
uncertainty."""

        # Kernel vector allows for evaluation of At. Env.
        k_v = self.get_kernel_vector(x_t, d)

        # Guarantee that alpha is up to date with training set
        if self.alpha is None or 3 * len(self.training_data) != len(
                self.alpha):
            self.set_L_alpha()

        # get predictive mean
        pred_mean = np.matmul(k_v, self.alpha)

        # get predictive variance without cholesky (possibly faster)
        self_kern = self.kernel(x_t, x_t, d, d, self.hyps,
                                self.cutoffs)
        pred_var = self_kern - \
            np.matmul(np.matmul(k_v, self.ky_mat_inv), k_v)

        return pred_mean, pred_var

    def predict_local_energy(self, x_t: AtomicEnvironment) -> float:
        """Predict the local energy of an atomic environment.

        :param x_t: Atomic environment of test atom.
        :type x_t: AtomicEnvironment
        :return: local energy in eV (up to a constant).
        :rtype: float
        """

        k_v = self.en_kern_vec(x_t)
        pred_mean = np.matmul(k_v, self.alpha)

        return pred_mean

    def predict_local_energy_and_var(self, x_t: AtomicEnvironment):
        """Predict the local energy of an atomic environment and its \
uncertainty."""

        # get kernel vector
        k_v = self.en_kern_vec(x_t)

        # get predictive mean
        pred_mean = np.matmul(k_v, self.alpha)

        # get predictive variance
        v_vec = solve_triangular(self.l_mat, k_v, lower=True)
        self_kern = self.energy_kernel(x_t, x_t, self.hyps,
                                       self.cutoffs)
        pred_var = self_kern - np.matmul(v_vec, v_vec)

        return pred_mean, pred_var

    def get_kernel_vector(self, x: AtomicEnvironment,
                          d_1: int) -> np.ndarray:
        """
        Compute kernel vector, comparing input environment to all environments
        in the GP's training set.

        :param x: data point to compare against kernel matrix
        :type x: AtomicEnvironment
        :param d_1: Cartesian component of force vector to get (1=x,2=y,3=z)
        :type d_1: int
        :return: kernel vector
        :rtype: np.ndarray
        """

        ds = [1, 2, 3]
        size = len(self.training_data) * 3
        k_v = np.zeros(size, )

        for m_index in range(size):
            x_2 = self.training_data[int(math.floor(m_index / 3))]
            d_2 = ds[m_index % 3]
            k_v[m_index] = self.kernel(x, x_2, d_1, d_2,
                                       self.hyps, self.cutoffs)
        return k_v

    def en_kern_vec(self, x: AtomicEnvironment) -> np.ndarray:
        """Compute the vector of energy/force kernels between an atomic \
environment and the environments in the training set."""

        ds = [1, 2, 3]
        size = len(self.training_data) * 3
        k_v = np.zeros(size, )

        for m_index in range(size):
            x_2 = self.training_data[int(math.floor(m_index / 3))]
            d_2 = ds[m_index % 3]
            k_v[m_index] = self.energy_force_kernel(x_2, x, d_2,
                                                    self.hyps, self.cutoffs)

        return k_v

    def set_L_alpha(self):
        """
        Invert the covariance matrix, setting L (a lower triangular
        matrix s.t. L L^T = (K + sig_n^2 I) ) and alpha, the inverse
        covariance matrix multiplied by the vector of training labels.
        The forces and variances are later obtained using alpha.
        :return:
        """
        if self.par:
            hyp_mat, ky_mat = \
                get_ky_and_hyp_par(self.hyps, self.training_data,
                                   self.training_labels_np,
                                   self.kernel_grad, self.cutoffs)
        else:
            hyp_mat, ky_mat = \
                get_ky_and_hyp(self.hyps, self.training_data,
                               self.training_labels_np,
                               self.kernel_grad, self.cutoffs)

        like, like_grad = \
            get_like_grad_from_mats(ky_mat, hyp_mat, self.training_labels_np)
        l_mat = np.linalg.cholesky(ky_mat)
        l_mat_inv = np.linalg.inv(l_mat)
        ky_mat_inv = l_mat_inv.T @ l_mat_inv
        alpha = np.matmul(ky_mat_inv, self.training_labels_np)

        self.ky_mat = ky_mat
        self.l_mat = l_mat
        self.alpha = alpha
        self.ky_mat_inv = ky_mat_inv
        self.l_mat_inv = l_mat_inv

        self.likelihood = like
        self.likelihood_gradient = like_grad

    def update_L_alpha(self):
        """
        Update the GP's L matrix and alpha vector.
        """

        # Set L matrix and alpha if set_L_alpha has not been called yet
        if self.l_mat is None:
            self.set_L_alpha()
            return

        n = self.l_mat_inv.shape[0]
        N = len(self.training_data)
        m = N - n // 3  # number of new data added
        ky_mat = np.zeros((3 * N, 3 * N))
        ky_mat[:n, :n] = self.ky_mat
        # calculate kernels for all added data
        for i in range(m):
            ind = n // 3 + i
            x_t = self.training_data[ind]
            k_vi = np.array([self.get_kernel_vector(x_t, d + 1)
                             for d in range(3)]).T  # (n+3m) x 3
            ky_mat[:, 3 * ind:3 * ind + 3] = k_vi
            ky_mat[3 * ind:3 * ind + 3, :n] = k_vi[:n, :].T
        sigma_n = self.hyps[-1]
        ky_mat[n:, n:] += sigma_n ** 2 * np.eye(3 * m)

        l_mat = np.linalg.cholesky(ky_mat)
        l_mat_inv = np.linalg.inv(l_mat)
        ky_mat_inv = l_mat_inv.T @ l_mat_inv
        alpha = np.matmul(ky_mat_inv, self.training_labels_np)

        self.ky_mat = ky_mat
        self.l_mat = l_mat
        self.alpha = alpha
        self.ky_mat_inv = ky_mat_inv
        self.l_mat_inv = l_mat_inv

    def __str__(self):
        """String representation of the GP model."""

        thestr = "GaussianProcess Object\n"
        thestr += 'Kernel: {}\n'.format(self.kernel_name)
        thestr += "Training points: {}\n".format(len(self.training_data))
        thestr += 'Cutoffs: {}\n'.format(self.cutoffs)
        thestr += 'Model Likelihood: {}\n'.format(self.likelihood)

        thestr += 'Hyperparameters: \n'
        if self.hyp_labels is None:
            # Put unlabeled hyperparameters on one line
            thestr = thestr[:-1]
            thestr += str(self.hyps) + '\n'
        else:
            for hyp, label in zip(self.hyps, self.hyp_labels):
                thestr += "{}: {}\n".format(label, hyp)

        return thestr

    def as_dict(self):
        """Dictionary representation of the GP model."""

        out_dict = dict(vars(self))

        out_dict['training_data'] = [env.as_dict() for env in
                                     self.training_data]
        # Remove the callables
        del out_dict['kernel']
        del out_dict['kernel_grad']

        return out_dict

    @staticmethod
    def from_dict(dictionary):
        """Create GP object from dictionary representation."""

        if 'mc' in dictionary['kernel_name']:
            force_kernel, grad = \
                str_to_mc_kernel(dictionary['kernel_name'], include_grad=True)
        else:
            force_kernel, grad = str_to_kernel(dictionary['kernel_name'],
                                               include_grad=True)

        if dictionary['energy_kernel'] is not None:
            energy_kernel = str_to_kernel(dictionary['energy_kernel'])
        else:
            energy_kernel = None

        if dictionary['energy_force_kernel'] is not None:
            energy_force_kernel = \
                str_to_kernel(dictionary['energy_force_kernel'])
        else:
            energy_force_kernel = None

        new_gp = GaussianProcess(kernel=force_kernel,
                                 kernel_grad=grad,
                                 energy_kernel=energy_kernel,
                                 energy_force_kernel=energy_force_kernel,
                                 cutoffs=np.array(dictionary['cutoffs']),
                                 hyps=np.array(dictionary['hyps']),
                                 hyp_labels=dictionary['hyp_labels'],
                                 par=dictionary['par'],
                                 maxiter=dictionary['maxiter'],
                                 opt_algorithm=dictionary['algo'])

        # Save time by attempting to load in computed attributes
        new_gp.l_mat = np.array(dictionary.get('l_mat', None))
        new_gp.l_mat_inv = np.array(dictionary.get('l_mat_inv', None))
        new_gp.alpha = np.array(dictionary.get('alpha', None))
        new_gp.ky_mat = np.array(dictionary.get('ky_mat', None))
        new_gp.ky_mat_inv = np.array(dictionary.get('ky_mat_inv', None))

        new_gp.training_data = [AtomicEnvironment.from_dict(env) for env in
                                dictionary['training_data']]
        new_gp.training_labels = dictionary['training_labels']

        new_gp.likelihood = dictionary['likelihood']
        new_gp.likelihood_gradient = dictionary['likelihood_gradient']

        return new_gp



import numpy as np
from math import exp
from numba import njit
import flare.cutoffs as cf


# -----------------------------------------------------------------------------
#                        two plus three body kernels
# -----------------------------------------------------------------------------


def two_plus_three_body(env1, env2, d1, d2, hyps, cutoffs,
                        cutoff_func=cf.quadratic_cutoff):

    two_term = two_body_jit(env1.bond_array_2, env2.bond_array_2,
                            d1, d2, hyps[0], hyps[1], cutoffs[0], cutoff_func)

    three_term = \
        three_body_jit(env1.bond_array_3, env2.bond_array_3,
                       env1.cross_bond_inds, env2.cross_bond_inds,
                       env1.cross_bond_dists, env2.cross_bond_dists,
                       env1.triplet_counts, env2.triplet_counts,
                       d1, d2, hyps[2], hyps[3], cutoffs[1], cutoff_func)
    return two_term + three_term


def two_plus_three_body_grad(env1, env2, d1, d2, hyps, cutoffs,
                             cutoff_func=cf.quadratic_cutoff):

    kern2, ls2, sig2 = \
        two_body_grad_jit(env1.bond_array_2, env2.bond_array_2,
                          d1, d2, hyps[0], hyps[1], cutoffs[0], cutoff_func)

    kern3, sig3, ls3 = \
        three_body_grad_jit(env1.bond_array_3, env2.bond_array_3,
                            env1.cross_bond_inds, env2.cross_bond_inds,
                            env1.cross_bond_dists, env2.cross_bond_dists,
                            env1.triplet_counts, env2.triplet_counts,
                            d1, d2, hyps[2], hyps[3], cutoffs[1], cutoff_func)

    return kern2 + kern3, np.array([sig2, ls2, sig3, ls3])


def two_plus_three_force_en(env1, env2, d1, hyps, cutoffs,
                            cutoff_func=cf.quadratic_cutoff):

    two_term = two_body_force_en_jit(env1.bond_array_2, env2.bond_array_2,
                                     d1, hyps[0], hyps[1], cutoffs[0],
                                     cutoff_func)/2

    three_term = \
        three_body_force_en_jit(env1.bond_array_3, env2.bond_array_3,
                                env1.cross_bond_inds, env2.cross_bond_inds,
                                env1.cross_bond_dists,
                                env2.cross_bond_dists,
                                env1.triplet_counts, env2.triplet_counts,
                                d1, hyps[2], hyps[3], cutoffs[1],
                                cutoff_func)/3

    return two_term + three_term


def two_plus_three_en(env1, env2, hyps, cutoffs,
                      cutoff_func=cf.quadratic_cutoff):

    two_term = two_body_en_jit(env1.bond_array_2, env2.bond_array_2,
                               hyps[0], hyps[1], cutoffs[0], cutoff_func)

    three_term = \
        three_body_en_jit(env1.bond_array_3, env2.bond_array_3,
                          env1.cross_bond_inds, env2.cross_bond_inds,
                          env1.cross_bond_dists, env2.cross_bond_dists,
                          env1.triplet_counts, env2.triplet_counts,
                          hyps[2], hyps[3], cutoffs[1], cutoff_func)

    return two_term + three_term


# -----------------------------------------------------------------------------
#                              two body kernels
# -----------------------------------------------------------------------------


def two_body(env1, env2, d1, d2, hyps, cutoffs,
             cutoff_func=cf.quadratic_cutoff):
    sig = hyps[0]
    ls = hyps[1]
    r_cut = cutoffs[0]

    return two_body_jit(env1.bond_array_2, env2.bond_array_2,
                        d1, d2, sig, ls, r_cut, cutoff_func)


def two_body_grad(env1, env2, d1, d2, hyps, cutoffs,
                  cutoff_func=cf.quadratic_cutoff):
    sig = hyps[0]
    ls = hyps[1]
    r_cut = cutoffs[0]

    kernel, ls_derv, sig_derv = \
        two_body_grad_jit(env1.bond_array_2, env2.bond_array_2,
                          d1, d2, sig, ls, r_cut, cutoff_func)
    kernel_grad = np.array([sig_derv, ls_derv])
    return kernel, kernel_grad


def two_body_force_en(env1, env2, d1, hyps, cutoffs,
                      cutoff_func=cf.quadratic_cutoff):
    sig = hyps[0]
    ls = hyps[1]
    r_cut = cutoffs[0]

    # divide by two to account for double counting
    return two_body_force_en_jit(env1.bond_array_2, env2.bond_array_2,
                                 d1, sig, ls, r_cut, cutoff_func)/2


def two_body_en(env1, env2, hyps, cutoffs,
                cutoff_func=cf.quadratic_cutoff):
    sig = hyps[0]
    ls = hyps[1]
    r_cut = cutoffs[0]

    return two_body_en_jit(env1.bond_array_2, env2.bond_array_2,
                           sig, ls, r_cut, cutoff_func)


# -----------------------------------------------------------------------------
#                              three body kernels
# -----------------------------------------------------------------------------


def three_body(env1, env2, d1, d2, hyps, cutoffs,
               cutoff_func=cf.quadratic_cutoff):
    sig = hyps[0]
    ls = hyps[1]
    r_cut = cutoffs[1]

    return three_body_jit(env1.bond_array_3, env2.bond_array_3,
                          env1.cross_bond_inds, env2.cross_bond_inds,
                          env1.cross_bond_dists, env2.cross_bond_dists,
                          env1.triplet_counts, env2.triplet_counts,
                          d1, d2, sig, ls, r_cut, cutoff_func)


def three_body_grad(env1, env2, d1, d2, hyps, cutoffs,
                    cutoff_func=cf.quadratic_cutoff):
    sig = hyps[0]
    ls = hyps[1]
    r_cut = cutoffs[1]

    kernel, sig_derv, ls_derv = \
        three_body_grad_jit(env1.bond_array_3, env2.bond_array_3,
                            env1.cross_bond_inds, env2.cross_bond_inds,
                            env1.cross_bond_dists, env2.cross_bond_dists,
                            env1.triplet_counts, env2.triplet_counts,
                            d1, d2, sig, ls, r_cut, cutoff_func)

    kernel_grad = np.array([sig_derv, ls_derv])

    return kernel, kernel_grad


def three_body_force_en(env1, env2, d1, hyps, cutoffs,
                        cutoff_func=cf.quadratic_cutoff):
    sig = hyps[0]
    ls = hyps[1]
    r_cut = cutoffs[1]

    # divide by three to account for triple counting
    return three_body_force_en_jit(env1.bond_array_3, env2.bond_array_3,
                                   env1.cross_bond_inds, env2.cross_bond_inds,
                                   env1.cross_bond_dists,
                                   env2.cross_bond_dists,
                                   env1.triplet_counts, env2.triplet_counts,
                                   d1, sig, ls, r_cut, cutoff_func)/3


def three_body_en(env1, env2, hyps, cutoffs,
                  cutoff_func=cf.quadratic_cutoff):
    sig = hyps[0]
    ls = hyps[1]
    r_cut = cutoffs[1]

    return three_body_en_jit(env1.bond_array_3, env2.bond_array_3,
                             env1.cross_bond_inds, env2.cross_bond_inds,
                             env1.cross_bond_dists, env2.cross_bond_dists,
                             env1.triplet_counts, env2.triplet_counts,
                             sig, ls, r_cut, cutoff_func)


# -----------------------------------------------------------------------------
#                           two body numba functions
# -----------------------------------------------------------------------------


@njit
def two_body_jit(bond_array_1, bond_array_2, d1, d2, sig, ls,
                 r_cut, cutoff_func):
    kern = 0

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    sig2 = sig*sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        ci = bond_array_1[m, d1]
        fi, fdi = cutoff_func(r_cut, ri, ci)

        for n in range(bond_array_2.shape[0]):
            rj = bond_array_2[n, 0]
            cj = bond_array_2[n, d2]
            fj, fdj = cutoff_func(r_cut, rj, cj)
            r11 = ri - rj

            A = ci * cj
            B = r11 * ci
            C = r11 * cj
            D = r11 * r11

            kern += force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2,
                                 ls3, sig2)

    return kern


@njit
def two_body_grad_jit(bond_array_1, bond_array_2, d1, d2, sig, ls,
                      r_cut, cutoff_func):

    kern = 0
    sig_derv = 0
    ls_derv = 0

    sig2, sig3, ls1, ls2, ls3, ls4, ls5, ls6 = grad_constants(sig, ls)

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        ci = bond_array_1[m, d1]
        fi, fdi = cutoff_func(r_cut, ri, ci)

        for n in range(bond_array_2.shape[0]):
            rj = bond_array_2[n, 0]
            cj = bond_array_2[n, d2]
            fj, fdj = cutoff_func(r_cut, rj, cj)

            r11 = ri - rj

            A = ci * cj
            B = r11 * ci
            C = r11 * cj
            D = r11 * r11

            kern_term, sig_term, ls_term = \
                grad_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, ls4,
                            ls5, ls6, sig2, sig3)

            kern += kern_term
            sig_derv += sig_term
            ls_derv += ls_term

    return kern, ls_derv, sig_derv


@njit
def two_body_force_en_jit(bond_array_1, bond_array_2, d1, sig, ls, r_cut,
                          cutoff_func):
    kern = 0

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    sig2 = sig*sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        ci = bond_array_1[m, d1]
        fi, fdi = cutoff_func(r_cut, ri, ci)

        for n in range(bond_array_2.shape[0]):
            rj = bond_array_2[n, 0]
            fj, _ = cutoff_func(r_cut, rj, 0)

            r11 = ri - rj
            B = r11 * ci
            D = r11 * r11
            kern += force_energy_helper(B, D, fi, fj, fdi, ls1, ls2, sig2)

    return kern


@njit
def two_body_en_jit(bond_array_1, bond_array_2, sig, ls, r_cut, cutoff_func):
    kern = 0

    ls1 = 1 / (2 * ls * ls)
    sig2 = sig * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        fi, _ = cutoff_func(r_cut, ri, 0)

        for n in range(bond_array_2.shape[0]):
            rj = bond_array_2[n, 0]
            fj, _ = cutoff_func(r_cut, rj, 0)
            r11 = ri - rj
            kern += fi * fj * sig2 * exp(-r11 * r11 * ls1)

    return kern


# -----------------------------------------------------------------------------
#                           three body numba functions
# -----------------------------------------------------------------------------


@njit
def three_body_jit(bond_array_1, bond_array_2,
                   cross_bond_inds_1, cross_bond_inds_2,
                   cross_bond_dists_1, cross_bond_dists_2,
                   triplets_1, triplets_2,
                   d1, d2, sig, ls, r_cut, cutoff_func):
    kern = 0

    # pre-compute constants that appear in the inner loop
    sig2 = sig*sig
    ls1 = 1 / (2*ls*ls)
    ls2 = 1 / (ls*ls)
    ls3 = ls2*ls2

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ci1 = bond_array_1[m, d1]
        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m+n+1]
            ri2 = bond_array_1[ind1, 0]
            ci2 = bond_array_1[ind1, d1]
            fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)

            ri3 = cross_bond_dists_1[m, m+n+1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)

            fi = fi1*fi2*fi3
            fdi = fdi1*fi2*fi3+fi1*fdi2*fi3

            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                cj1 = bond_array_2[p, d2]
                fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)

                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p+1+q]
                    rj2 = bond_array_2[ind2, 0]
                    cj2 = bond_array_2[ind2, d2]
                    fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)

                    rj3 = cross_bond_dists_2[p, p+1+q]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)

                    fj = fj1*fj2*fj3
                    fdj = fdj1*fj2*fj3+fj1*fdj2*fj3
                    
                    kern += triplet_kernel(ci1, ci2, cj1, cj2, ri1, ri2, ri3,
                                           rj1, rj2, rj3, fi, fj, fdi, fdj,
                                           ls1, ls2, ls3, sig2)
    return kern


@njit
def three_body_grad_jit(bond_array_1, bond_array_2,
                        cross_bond_inds_1, cross_bond_inds_2,
                        cross_bond_dists_1, cross_bond_dists_2,
                        triplets_1, triplets_2,
                        d1, d2, sig, ls, r_cut, cutoff_func):
    """Kernel gradient for 3-body force comparisons."""

    kern = 0
    sig_derv = 0
    ls_derv = 0

    # pre-compute constants that appear in the inner loop
    sig2, sig3, ls1, ls2, ls3, ls4, ls5, ls6 = grad_constants(sig, ls)

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ci1 = bond_array_1[m, d1]
        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m+n+1]
            ri3 = cross_bond_dists_1[m, m+n+1]
            ri2 = bond_array_1[ind1, 0]
            ci2 = bond_array_1[ind1, d1]

            fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
            fi3, _ = cutoff_func(r_cut, ri3, 0)

            fi = fi1*fi2*fi3
            fdi = fdi1*fi2*fi3+fi1*fdi2*fi3

            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                cj1 = bond_array_2[p, d2]
                fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)

                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p+q+1]
                    rj3 = cross_bond_dists_2[p, p+q+1]
                    rj2 = bond_array_2[ind2, 0]
                    cj2 = bond_array_2[ind2, d2]

                    fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)
                    fj3, _ = cutoff_func(r_cut, rj3, 0)

                    fj = fj1*fj2*fj3
                    fdj = fdj1*fj2*fj3+fj1*fdj2*fj3

                    N, O, X = \
                        triplet_kernel_grad(ci1, ci2, cj1, cj2, ri1, ri2, ri3,
                                            rj1, rj2, rj3, fi, fj, fdi, fdj,
                                            ls1, ls2, ls3, ls4, ls5, ls6, sig2,
                                            sig3)

                    kern += N
                    sig_derv += O
                    ls_derv += X

    return kern, sig_derv, ls_derv


@njit
def three_body_force_en_jit(bond_array_1, bond_array_2,
                            cross_bond_inds_1,
                            cross_bond_inds_2,
                            cross_bond_dists_1,
                            cross_bond_dists_2,
                            triplets_1, triplets_2,
                            d1, sig, ls, r_cut, cutoff_func):
    """Kernel for 3-body force/energy comparisons."""

    kern = 0

    # pre-compute constants that appear in the inner loop
    sig2 = sig*sig
    ls1 = 1 / (2*ls*ls)
    ls2 = 1 / (ls*ls)

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ci1 = bond_array_1[m, d1]
        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m+n+1]
            ri2 = bond_array_1[ind1, 0]
            ci2 = bond_array_1[ind1, d1]
            fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
            ri3 = cross_bond_dists_1[m, m+n+1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)
            fi = fi1*fi2*fi3
            fdi = fdi1*fi2*fi3+fi1*fdi2*fi3

            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                fj1, _ = cutoff_func(r_cut, rj1, 0)

                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p+q+1]
                    rj2 = bond_array_2[ind2, 0]
                    fj2, _ = cutoff_func(r_cut, rj2, 0)
                    rj3 = cross_bond_dists_2[p, p+q+1]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    fj = fj1*fj2*fj3

                    kern += triplet_force_en_kernel(ci1, ci2, ri1, ri2, ri3,
                                                    rj1, rj2, rj3, fi, fj, fdi,
                                                    ls1, ls2, sig2)

    return kern


@njit
def three_body_en_jit(bond_array_1, bond_array_2,
                      cross_bond_inds_1,
                      cross_bond_inds_2,
                      cross_bond_dists_1,
                      cross_bond_dists_2,
                      triplets_1, triplets_2,
                      sig, ls, r_cut, cutoff_func):
    kern = 0

    sig2 = sig*sig
    ls2 = 1 / (2*ls*ls)

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        fi1, _ = cutoff_func(r_cut, ri1, 0)

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            fi2, _ = cutoff_func(r_cut, ri2, 0)

            ri3 = cross_bond_dists_1[m, m + n + 1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)
            fi = fi1*fi2*fi3

            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                fj1, _ = cutoff_func(r_cut, rj1, 0)

                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p + q + 1]
                    rj2 = bond_array_2[ind2, 0]
                    fj2, _ = cutoff_func(r_cut, rj2, 0)

                    rj3 = cross_bond_dists_2[p, p + q + 1]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    fj = fj1*fj2*fj3

                    r11 = ri1-rj1
                    r12 = ri1-rj2
                    r13 = ri1-rj3
                    r21 = ri2-rj1
                    r22 = ri2-rj2
                    r23 = ri2-rj3
                    r31 = ri3-rj1
                    r32 = ri3-rj2
                    r33 = ri3-rj3

                    C1 = r11*r11+r22*r22+r33*r33
                    C2 = r11*r11+r23*r23+r32*r32
                    C3 = r12*r12+r21*r21+r33*r33
                    C4 = r12*r12+r23*r23+r31*r31
                    C5 = r13*r13+r21*r21+r32*r32
                    C6 = r13*r13+r22*r22+r31*r31

                    k = exp(-C1*ls2)+exp(-C2*ls2)+exp(-C3*ls2)+exp(-C4*ls2) + \
                        exp(-C5*ls2)+exp(-C6*ls2)

                    kern += sig2*k*fi*fj

    return kern


# -----------------------------------------------------------------------------
#                            general helper functions
# -----------------------------------------------------------------------------


@njit
def grad_constants(sig, ls):
    sig2 = sig * sig
    sig3 = 2 * sig

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    ls4 = 1 / (ls * ls * ls)
    ls5 = ls * ls
    ls6 = ls2 * ls4

    return sig2, sig3, ls1, ls2, ls3, ls4, ls5, ls6


@njit
def force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2):
    E = exp(-D * ls1)
    F = E * B * ls2
    G = -E * C * ls2
    H = A * E * ls2 - B * C * E * ls3
    I = E * fdi * fdj
    J = F * fi * fdj
    K = G * fdi * fj
    L = H * fi * fj
    M = sig2 * (I + J + K + L)

    return M


@njit
def grad_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, ls4, ls5, ls6,
                sig2, sig3):
    E = exp(-D * ls1)
    F = E * B * ls2
    G = -E * C * ls2
    H = A * E * ls2 - B * C * E * ls3
    I = E * fdi * fdj
    J = F * fi * fdj
    K = G * fdi * fj
    L = H * fi * fj
    M = I + J + K + L
    N = sig2 * M
    O = sig3 * M
    P = E * D * ls4
    Q = B * (ls2 * P - 2 * E * ls4)
    R = -C * (ls2 * P - 2 * E * ls4)
    S = (A * ls5 - B * C) * (P * ls3 - 4 * E * ls6) + 2 * E * A * ls4
    T = P * fdi * fdj
    U = Q * fi * fdj
    V = R * fdi * fj
    W = S * fi * fj
    X = sig2 * (T + U + V + W)

    return N, O, X


@njit
def force_energy_helper(B, D, fi, fj, fdi, ls1, ls2, sig2):
    E = exp(-D * ls1)
    F = E * B * ls2
    G = -F * fi * fj
    H = -E * fdi * fj
    I = sig2 * (G + H)

    return I


# -----------------------------------------------------------------------------
#                        three body helper functions
# -----------------------------------------------------------------------------


@njit
def triplet_kernel(ci1, ci2, cj1, cj2, ri1, ri2, ri3, rj1, rj2, rj3, fi, fj,
                   fdi, fdj, ls1, ls2, ls3, sig2):
    r11 = ri1-rj1
    r12 = ri1-rj2
    r13 = ri1-rj3
    r21 = ri2-rj1
    r22 = ri2-rj2
    r23 = ri2-rj3
    r31 = ri3-rj1
    r32 = ri3-rj2
    r33 = ri3-rj3

    # sum over all six permutations
    M1 = three_body_helper_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)
    M2 = three_body_helper_2(ci2, ci1, cj2, cj1, r21, r13, r32, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)
    M3 = three_body_helper_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)
    M4 = three_body_helper_1(ci1, ci2, cj2, cj1, r12, r21, r33, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)
    M5 = three_body_helper_2(ci2, ci1, cj1, cj2, r22, r13, r31, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)
    M6 = three_body_helper_2(ci1, ci2, cj2, cj1, r11, r23, r32, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)

    return M1 + M2 + M3 + M4 + M5 + M6


@njit
def triplet_kernel_grad(ci1, ci2, cj1, cj2, ri1, ri2, ri3, rj1, rj2, rj3, fi,
                        fj, fdi, fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2,
                        sig3):
    r11 = ri1-rj1
    r12 = ri1-rj2
    r13 = ri1-rj3
    r21 = ri2-rj1
    r22 = ri2-rj2
    r23 = ri2-rj3
    r31 = ri3-rj1
    r32 = ri3-rj2
    r33 = ri3-rj3

    N1, O1, X1 = \
        three_body_grad_helper_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi, fj,
                                 fdi, fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2,
                                 sig3)
    N2, O2, X2 = \
        three_body_grad_helper_2(ci2, ci1, cj2, cj1, r21, r13, r32, fi, fj,
                                 fdi, fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2,
                                 sig3)
    N3, O3, X3 = \
        three_body_grad_helper_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi, fj,
                                 fdi, fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2,
                                 sig3)
    N4, O4, X4 = \
        three_body_grad_helper_1(ci1, ci2, cj2, cj1, r12, r21, r33, fi, fj,
                                 fdi, fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2,
                                 sig3)
    N5, O5, X5 = \
        three_body_grad_helper_2(ci2, ci1, cj1, cj2, r22, r13, r31, fi, fj,
                                 fdi, fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2,
                                 sig3)
    N6, O6, X6 = \
        three_body_grad_helper_2(ci1, ci2, cj2, cj1, r11, r23, r32, fi, fj,
                                 fdi, fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2,
                                 sig3)
    N = N1 + N2 + N3 + N4 + N5 + N6
    O = O1 + O2 + O3 + O4 + O5 + O6
    X = X1 + X2 + X3 + X4 + X5 + X6
    return N, O, X


@njit
def three_body_helper_1(ci1, ci2, cj1, cj2, r11, r22, r33,
                        fi, fj, fdi, fdj,
                        ls1, ls2, ls3, sig2):
    A = ci1*cj1+ci2*cj2
    B = r11*ci1+r22*ci2
    C = r11*cj1+r22*cj2
    D = r11*r11+r22*r22+r33*r33

    M = force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2)

    return M


@njit
def three_body_helper_2(ci1, ci2, cj1, cj2, r12, r23, r31,
                        fi, fj, fdi, fdj,
                        ls1, ls2, ls3, sig2):
    A = ci1*cj2
    B = r12*ci1+r23*ci2
    C = r12*cj2+r31*cj1
    D = r12*r12+r23*r23+r31*r31

    M = force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2)

    return M


@njit
def three_body_grad_helper_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2, sig3):
    A = ci1*cj1+ci2*cj2
    B = r11*ci1+r22*ci2
    C = r11*cj1+r22*cj2
    D = r11*r11+r22*r22+r33*r33

    N, O, X = grad_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, ls4,
                          ls5, ls6, sig2, sig3)

    return N, O, X


@njit
def three_body_grad_helper_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2, sig3):
    A = ci1*cj2
    B = r12*ci1+r23*ci2
    C = r12*cj2+r31*cj1
    D = r12*r12+r23*r23+r31*r31

    N, O, X = grad_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, ls4,
                          ls5, ls6, sig2, sig3)

    return N, O, X


@njit
def three_body_en_helper(ci1, ci2, r11, r22, r33, fi, fj, fdi, ls1, ls2, sig2):
    B = r11 * ci1 + r22 * ci2
    D = r11 * r11 + r22 * r22 + r33 * r33

    return force_energy_helper(B, D, fi, fj, fdi, ls1, ls2, sig2)


@njit
def triplet_force_en_kernel(ci1, ci2, ri1, ri2, ri3, rj1, rj2, rj3,
                            fi, fj, fdi, ls1, ls2, sig2):
    r11 = ri1-rj1
    r12 = ri1-rj2
    r13 = ri1-rj3
    r21 = ri2-rj1
    r22 = ri2-rj2
    r23 = ri2-rj3
    r31 = ri3-rj1
    r32 = ri3-rj2
    r33 = ri3-rj3

    I1 = three_body_en_helper(ci1, ci2, r11, r22, r33, fi, fj,
                              fdi, ls1, ls2, sig2)
    I2 = three_body_en_helper(ci1, ci2, r13, r21, r32, fi, fj,
                              fdi, ls1, ls2, sig2)
    I3 = three_body_en_helper(ci1, ci2, r12, r23, r31, fi, fj,
                              fdi, ls1, ls2, sig2)
    I4 = three_body_en_helper(ci1, ci2, r12, r21, r33, fi, fj,
                              fdi, ls1, ls2, sig2)
    I5 = three_body_en_helper(ci1, ci2, r13, r22, r31, fi, fj,
                              fdi, ls1, ls2, sig2)
    I6 = three_body_en_helper(ci1, ci2, r11, r23, r32, fi, fj,
                              fdi, ls1, ls2, sig2)

    return I1 + I2 + I3 + I4 + I5 + I6


if __name__ == '__main__':
    import pytest
    import numpy as np
    import sys
    from random import random, randint
    from copy import deepcopy
    sys.path.append('../otf_engine')
    import env
    import gp
    import struc
    import kernels as en

    # create env 1
    delt = 1e-5
    cell = np.eye(3)
    cutoff = 1
    cutoffs = np.array([1, 1])

    positions_1 = [np.array([0., 0., 0.]),
                   np.array([random(), random(), random()]),
                   np.array([random(), random(), random()])]
    positions_2 = deepcopy(positions_1)
    positions_2[0][0] = delt

    positions_3 = deepcopy(positions_1)
    positions_3[0][0] = -delt

    species_1 = ['A', 'B', 'A']
    atom_1 = 0
    test_structure_1 = struc.Structure(cell, species_1, positions_1)
    test_structure_2 = struc.Structure(cell, species_1, positions_2)
    test_structure_3 = struc.Structure(cell, species_1, positions_3)

    env1_1 = env.AtomicEnvironment(test_structure_1, atom_1, cutoffs)
    env1_2 = env.AtomicEnvironment(test_structure_2, atom_1, cutoffs)
    env1_3 = env.AtomicEnvironment(test_structure_3, atom_1, cutoffs)

    # create env 2
    positions_1 = [np.array([0., 0., 0.]),
                   np.array([random(), random(), random()]),
                   np.array([random(), random(), random()])]
    positions_2 = deepcopy(positions_1)
    positions_2[0][1] = delt
    positions_3 = deepcopy(positions_1)
    positions_3[0][1] = -delt

    species_2 = ['A', 'A', 'B']
    atom_2 = 0
    test_structure_1 = struc.Structure(cell, species_2, positions_1)
    test_structure_2 = struc.Structure(cell, species_2, positions_2)
    test_structure_3 = struc.Structure(cell, species_2, positions_3)

    env2_1 = env.AtomicEnvironment(test_structure_1, atom_2, cutoffs)
    env2_2 = env.AtomicEnvironment(test_structure_2, atom_2, cutoffs)
    env2_3 = env.AtomicEnvironment(test_structure_3, atom_2, cutoffs)

    sig = 1
    ls = 0.1
    d1 = 1
    d2 = 2

    hyps = np.array([sig, ls])

    # check force kernel
    calc1 = en.two_body_en(env1_2, env2_2, hyps, cutoffs)
    calc2 = en.two_body_en(env1_3, env2_3, hyps, cutoffs)
    calc3 = en.two_body_en(env1_2, env2_3, hyps, cutoffs)
    calc4 = en.two_body_en(env1_3, env2_2, hyps, cutoffs)

    kern_finite_diff = (calc1 + calc2 - calc3 - calc4) / (4*delt**2)
    kern_analytical = en.two_body(env1_1, env2_1,
                                  d1, d2, hyps, cutoffs)

    tol = 1e-4

    print(kern_finite_diff)
    print(kern_analytical)
    assert(np.isclose(kern_finite_diff, kern_analytical, atol=tol))

_str_to_kernel = {'two_body': two_body,
                  'two_body_en': two_body_en,
                  'two_body_force_en': two_body_force_en,
                  'three_body': three_body,
                  'three_body_en': three_body_en,
                  'three_body_force_en': three_body_force_en,
                  'two_plus_three_body': two_plus_three_body,
                  'two_plus_three_en': two_plus_three_en,
                  'two_plus_three_force_en': two_plus_three_force_en
                  }


def str_to_kernel(string: str, include_grad: bool=False):

    if string not in _str_to_kernel.keys():
        raise ValueError("Kernel {} not found in list of available "
                         "kernels{}:".format(string,_str_to_kernel.keys()))

    if not include_grad:
        return _str_to_kernel[string]
    else:
        if 'two' in string and 'three' in string:
            return _str_to_kernel[string], two_plus_three_body_grad
        elif 'two' in string and 'three' not in string:
            return _str_to_kernel[string], two_body_grad
        elif 'two' not in string and 'three' in string:
            return _str_to_kernel[string], three_body_grad
        else:
            raise ValueError("Gradient callable for {} not found".format(
                string))



import sys
import numpy as np
import datetime
import time
from typing import List
import copy
import multiprocessing as mp
import concurrent.futures
from flare import struc, gp, env, qe_util, md, output
from flare.mgp.mgp import MappedGaussianProcess


class Validate:
    def __init__(self, qe_input: str, dt: float, number_of_steps: int,
                 dft_steps, mgp_model: MappedGaussianProcess, pw_loc: str,
                 prev_pos_init: np.ndarray=None, par: bool=False, skip: int=0,
                 calculate_energy=False, output_name='validate.out', no_cpus=1):

        self.qe_input = qe_input
        self.dt = dt
        self.number_of_steps = number_of_steps
        self.mgp_model = mgp_model
        self.gp = mgp_model.GP
        self.pw_loc = pw_loc
        self.skip = skip
        self.dft_step = True
        self.dft_steps = dft_steps

        # parse input file
        positions, species, cell, masses = \
            qe_util.parse_qe_input(self.qe_input)

        self.structure = struc.Structure(cell=cell, species=species,
                                         positions=positions,
                                         mass_dict=masses,
                                         prev_positions=prev_pos_init)

        self.noa = self.structure.positions.shape[0]
        self.atom_list = list(range(self.noa))
        self.curr_step = 0

        # initialize local energies
        if calculate_energy:
            self.local_energies = np.zeros(self.noa)
        else:
            self.local_energies = None

        self.dft_count = 0

        # set pred function
        if not par and not calculate_energy:
            self.pred_func = self.predict_on_structure_mgp
        elif par and not calculate_energy:
            self.pred_func = self.predict_on_structure_par
        elif not par and calculate_energy:
            self.pred_func = self.predict_on_structure_en
        elif par and calculate_energy:
            self.pred_func = self.predict_on_structure_par_en
        self.par = par

        self.output_name = output_name

        # set number of cpus for qe runs
        self.no_cpus = no_cpus

    def write_mgp_header(self):
        output.write_to_output('bounds 2:'+str(self.mgp_model.bounds_2), self.output_name)
        output.write_to_output('bounds 3:'+str(self.mgp_model.bounds_2), self.output_name)
        output.write_to_output('grid num 2:'+str(self.mgp_model.grid_num_2), self.output_name)
        output.write_to_output('grid num 3:'+str(self.mgp_model.grid_num_3), self.output_name)

    def run(self):
        output.write_header(self.gp.cutoffs, self.gp.kernel_name, self.gp.hyps,
                            self.gp.algo, self.dt, self.number_of_steps,
                            self.structure, self.output_name, 1.)
        self.write_mgp_header()

        counter = 0
        self.start_time = time.time()

        while self.curr_step < self.number_of_steps:
            self.pred_func()
            self.dft_step = False
            new_pos = md.update_positions(self.dt, self.noa,
                                          self.structure)

            if self.curr_step in self.dft_steps:
                # record GP forces
                self.update_temperature(new_pos)
                self.record_state()

                # run DFT and record forces
                self.dft_step = True
                self.run_dft()
                new_pos = md.update_positions(self.dt, self.noa,
                                              self.structure)
                self.update_temperature(new_pos)
                self.record_state()

            # write gp forces
            if counter >= self.skip and not self.dft_step:
                self.update_temperature(new_pos)
                self.record_state()
                counter = 0

            counter += 1
            self.update_positions(new_pos)
            self.curr_step += 1

        output.conclude_run(self.output_name)

    def predict_on_structure_mgp(self): # changed
        """
        Assign forces to self.structure based on self.gp
        """

        output.write_to_output('\npredict with mapping:\n', self.output_name)
        for n in range(self.structure.nat):
            chemenv = env.AtomicEnvironment(self.structure, n, self.gp.cutoffs)
            force, var = self.mgp_model.predict(chemenv)
            self.structure.forces[n][:] = force
            self.structure.stds[n][:] = np.sqrt(np.absolute(var))
        self.structure.dft_forces = False


    def run_dft(self):
        output.write_to_output('\nCalling Quantum Espresso...\n',
                               self.output_name)

        # calculate DFT forces
        forces = qe_util.run_espresso_par(self.qe_input, self.structure,
                                          self.pw_loc, self.no_cpus)
        self.structure.forces = forces

        # write wall time of DFT calculation
        self.dft_count += 1
        output.write_to_output('QE run complete.\n', self.output_name)
        time_curr = time.time() - self.start_time
        output.write_to_output('number of DFT calls: %i \n' % self.dft_count,
                               self.output_name)
        output.write_to_output('wall time from start: %.2f s \n' % time_curr,
                               self.output_name)

    def update_positions(self, new_pos):
        self.structure.prev_positions = self.structure.positions
        self.structure.positions = new_pos
        self.structure.wrap_positions()

    def update_temperature(self, new_pos):
        KE, temperature, velocities = \
                md.calculate_temperature(new_pos, self.structure, self.dt,
                                         self.noa)
        self.KE = KE
        self.temperature = temperature
        self.velocities = velocities

    def record_state(self):
        output.write_md_config(self.dt, self.curr_step, self.structure,
                               self.temperature, self.KE,
                               self.local_energies, self.start_time,
                               self.output_name, self.dft_step,
                               self.velocities)



import time
import math
from math import exp
import numpy as np
from numba import njit
from scipy.linalg import solve_triangular
import multiprocessing as mp
import sys
sys.path.append('../../flare/')
from memory_profiler import profile
import subprocess
import struct

import flare.gp as gp
import flare.env as env
from flare.kernels import two_body, three_body, two_plus_three_body, two_body_jit
import flare.struc as struc

import flare.mgp.utils as utils
from flare.mgp.splines_methods import PCASplines, SplinesInterpolation

class MappedGaussianProcess:
    
    def __init__(self, GP, grid_params, struc_params):
    
        '''
        param: struc_params = {'species': 'C', 'cube_lat': 2*1.763391008}
        param: grid_params = {'grid_num': list, 'bounds': list, 
                            'svd_rank': int>0, 'load_grid': None, 
                            'load_svd': None}
        '''
        self.GP = GP
        self.grid_params = grid_params
        self.struc_params = struc_params
        self.bodies = str(grid_params['bodies'])
        self.grid_num_2 = grid_params['grid_num_2']
        self.bounds_2 = grid_params['bounds_2']
        self.grid_num_3 = grid_params['grid_num_3']
        self.bounds_3 = grid_params['bounds_3']
        self.svd_rank_2 = grid_params['svd_rank_2']
        self.svd_rank_3 = grid_params['svd_rank_3']
        self.update = grid_params['update']
        
        bond_struc = self.build_bond_struc(struc_params)
        if len(GP.training_data) > 0:
            if self.bodies == '2':
                self.map = Map2body(self.grid_num_2, self.bounds_2, self.GP, bond_struc,  
                       self.bodies, grid_params['load_grid'], self.svd_rank_2)
            elif self.bodies == '3':
                self.map = Map3body(self.grid_num_3, self.bounds_3, self.GP, bond_struc, 
                       self.bodies, grid_params['load_grid'], 
                       grid_params['load_svd'], self.svd_rank_3, self.update)
            elif self.bodies == '2+3':
                self.map_2 = Map2body(self.grid_num_2, self.bounds_2, self.GP, bond_struc[0],
                         self.bodies, grid_params['load_grid'], self.svd_rank_2)
                self.map_3 = Map3body(self.grid_num_3, self.bounds_3, self.GP,
                         bond_struc[1], self.bodies, grid_params['load_grid'],
                         grid_params['load_svd'], self.svd_rank_3, self.update)

    def build_bond_struc(self, struc_params):
    
        '''
        build a bond structure, used in grid generating
        '''
    
        cutoff = np.min(self.GP.cutoffs)
        cell = struc_params['cube_lat']
        mass_dict = struc_params['mass_dict']
        bond_struc = []
        for bodies in [2, 3]:
            species = [struc_params['species'] for i in range(bodies)]
            positions = [[(i+1)/(bodies+1)*cutoff, 0, 0] \
                        for i in range(bodies)]
            bond_struc.append(struc.Structure(cell, species, positions, mass_dict))
        if self.bodies == '2':
            return bond_struc[0]
        elif self.bodies == '3':
            return bond_struc[1]
        elif self.bodies == '2+3':
            return bond_struc

    def predict(self, atom_env, mean_only=False):
        if self.bodies == '2':
            f, v = self.map.predict(atom_env, self.GP, mean_only)
        elif self.bodies == '3':
            f, v = self.map.predict(atom_env, self.GP, mean_only)
        elif self.bodies == '2+3':
            f2, kern2, v2 = self.map_2.predict(atom_env, self.GP, mean_only)
            f3, kern3, v3 = self.map_3.predict(atom_env, self.GP, mean_only)
            f = f2 + f3
            v = kern2 + kern3 - np.sum((v2 + v3)**2, axis=0)
        return f, v
              
class Map2body:
   
    def __init__(self, grid_num, bounds, GP, bond_struc, bodies='2', load_prefix=None, svd_rank=0): 
    
        '''
        param grids: the 1st element is the number of grids for mean prediction, 
                    the 2nd is for var
        '''       
        
        self.grid_num = grid_num
        self.l_bound, self.u_bound = bounds
        self.cutoffs = GP.cutoffs
        self.bodies = bodies
        self.svd_rank = svd_rank
        
        if self.bodies == '2':
            y_mean, y_var = self.GenGrid(GP, bond_struc)
        elif self.bodies == '2+3':
            y_mean, y_var = self.GenGrid_svd(GP, bond_struc)

        self.build_map(y_mean, y_var)

    def GenGrid(self, GP, bond_struc, processes=mp.cpu_count()):
    
        '''
        generate grid data of mean prediction and L^{-1}k* for each triplet
        default implemented in a parallelized style
        '''
        processes = mp.cpu_count()
        nop = self.grid_num
        bond_lengths = np.linspace(self.l_bound, self.u_bound, nop)
        bond_means = np.zeros([nop])
        bond_vars = np.zeros([nop, nop])

        env1 = env.AtomicEnvironment(bond_struc, 0, self.cutoffs)
        env2 = env.AtomicEnvironment(bond_struc, 0, self.cutoffs)

        pool_list = [(i, bond_lengths[i], bond_lengths, GP, env1, env2) for i in range(nop)]
        pool = mp.Pool(processes=processes)
        A_list = pool.starmap(self._GenGrid_inner, pool_list)
        pool.close()
        pool.join()

        A_list.sort(key=lambda x: x[0])
        for b1 in range(nop):            
            bond_means[b1] = A_list[b1][1]
            bond_vars[b1, :] = A_list[b1][2]
        
        return bond_means, bond_vars


    def _GenGrid_inner(self, b1, r1, bond_lengths, GP, env1, env2):
    
        '''
        generate grid for each angle, used to parallelize grid generation
        '''
        
        
        nop = self.grid_num
        bond_vars = np.zeros(nop)
        
        bond1 = np.array([r1, 1.0, 0.0, 0.0])
        env1.bond_array_2 = np.array([bond1])
#        env1.cross_bond_dists = np.array([[0]])
        
        k1_v = GP.get_kernel_vector(env1, 1)
        v1_vec = solve_triangular(GP.l_mat, k1_v, lower=True)
        mean_diff = np.matmul(k1_v, GP.alpha)
        bond_means = mean_diff
        
        for b2, r2 in enumerate(bond_lengths):
            bond2 = np.array([r2, 1.0, 0.0, 0.0])
            env2.bond_array_2 = np.array([bond2])
#            env2.cross_bond_dists = np.array([[0]])
            
            k2_v = GP.get_kernel_vector(env2, 1)
            v2_vec = solve_triangular(GP.l_mat, k2_v, lower=True)
            self_kern = GP.kernel(env1, env2, 1, 1, GP.hyps, GP.cutoffs)
            var_diff = self_kern - np.matmul(v1_vec, v2_vec)
            bond_vars[b2] = var_diff   
                      
        return b1, bond_means, bond_vars

    #@profile
    def GenGrid_svd(self, GP, bond_struc, processes=mp.cpu_count()):
    
        '''
        generate grid data of mean prediction and L^{-1}k* for each triplet
         implemented in a parallelized style
        '''

        # ------ change GP kernel to 2 body ------
        GP.kernel = two_body
        original_cutoffs = np.copy(GP.cutoffs)
        GP.cutoffs = [GP.cutoffs[0]]
        original_hyps = np.copy(GP.hyps)
        GP.hyps = [GP.hyps[0], GP.hyps[1], GP.hyps[-1]]

        # ------ construct grids ------
        nop = self.grid_num
        bond_lengths = np.linspace(self.l_bound[0], self.u_bound[0], nop)
        bond_means = np.zeros([nop])
        bond_vars = np.zeros([nop, len(GP.alpha)])
        env12 = env.AtomicEnvironment(bond_struc, 0, self.cutoffs)
        
        pool_list = [(i, bond_lengths, GP, env12)\
                     for i in range(nop)]
        pool = mp.Pool(processes=processes)
        A_list = pool.map(self._GenGrid_svd_inner, pool_list)
        for p in range(nop):
            bond_means[p] = A_list[p][0]
            bond_vars[p, :] = A_list[p][1]
        pool.close()
        pool.join()

        # ------ change back original GP ------
        GP.cutoffs = original_cutoffs
        GP.hyps = original_hyps
        GP.kernel = two_plus_three_body
       
        return bond_means, bond_vars

    def _GenGrid_svd_inner(self, params):
    
        '''
        generate grid for each angle, used to parallelize grid generation
        '''
        b, bond_lengths, GP, env12 = params
        nop = self.grid_num
        r = bond_lengths[b]
        env12.bond_array_2 = np.array([[r, 1, 0, 0]])
        k12_v = GP.get_kernel_vector(env12, 1)   
        v12_vec = solve_triangular(GP.l_mat, k12_v, lower=True)
        mean_diff = np.matmul(k12_v, GP.alpha)
        bond_means = mean_diff
        bond_vars = v12_vec  
                      
        return bond_means, bond_vars


    def build_map(self, y_mean, y_var):
    
        '''
        build 1-d spline function for mean, 2-d for var
        '''
        
        self.mean = SplinesInterpolation(y_mean, 
                    u_bounds=np.array(self.u_bound), 
                    l_bounds=np.array(self.l_bound), 
                    orders=np.array([self.grid_num]))
        if self.bodies == '2':
            self.var = SplinesInterpolation(y_var, 
                        u_bounds=np.array([self.u_bound, self.u_bound]), 
                        l_bounds=np.array([self.l_bound, self.l_bound]), 
                        orders=np.array([self.grid_num, self.grid_num]))
        elif self.bodies == '2+3':
            self.var = PCASplines(y_var, u_bounds=np.array(self.u_bound), 
                       l_bounds=np.array(self.l_bound), 
                       orders=np.array([self.grid_num]), 
                       svd_rank=self.svd_rank, load_svd=None)


    def predict(self, atom_env, GP, mean_only):
    
        '''
        predict for an atom environment
        param: atom_env: ChemicalEnvironment
        return force on an atom with its variance
        '''
        
        bond_lengths = atom_env.bond_array_2[:,0]
        bond_dirs = atom_env.bond_array_2[:,1:]
        bond_num = len(bond_lengths)
       
        bond_lengths = np.expand_dims(bond_lengths, axis=1)
        mean_diffs = self.mean(bond_lengths)
        bond_forces = [mean_diffs*bond_dirs[:,i] for i in range(3)]
        atom_mean = np.sum(bond_forces, axis=1)
        
        atom_var = np.zeros(3)
        if not mean_only:
            if self.bodies == '2':
                ind_1, ind_2 = np.meshgrid(np.arange(bond_num), np.arange(bond_num))
                ind_1 = np.reshape(ind_1, (ind_1.shape[0]*ind_1.shape[1], 1)) 
                ind_2 = np.reshape(ind_2, (ind_2.shape[0]*ind_2.shape[1], 1)) 
                bond_1, bond_2 = (bond_lengths[ind_1], bond_lengths[ind_2])        
                bond_xyz1 = bond_dirs[ind_1,:] 
                bond_xyz2 = bond_dirs[ind_2,:] 
                bond_concat = np.concatenate([bond_1, bond_2], axis=1)  
                var_diffs = self.var(bond_concat)
                var_diffs = np.repeat(np.expand_dims(var_diffs, axis=1), 3, axis=1)
                atom_var = np.sum(var_diffs*bond_xyz1[:,0,:]*bond_xyz2[:,0,:], axis=0) 
                return atom_mean, atom_var    
            elif self.bodies == '2+3':
                sig_2, ls_2, sig_3, ls_3, noise = GP.hyps
                LambdaU = self.var(bond_lengths)
                VLambdaU = self.var.V @ LambdaU
                v = VLambdaU @ bond_dirs
                self_kern = np.zeros(3)
                for d in range(3):
                    self_kern[d] = self_two_body_jit(atom_env.bond_array_2, d+1, 
                           sig_2, ls_2, GP.cutoffs[0], quadratic_cutoff)
                return atom_mean, self_kern, v
        else:
            if self.bodies == '2':
                return atom_mean, atom_var
            elif self.bodies == '2+3':
                return atom_mean, 0, 0



class Map3body:
   
    def __init__(self, grid_num, bounds, GP, bond_struc, bodies='3', 
                load_grid=False, load_svd=None, svd_rank=0, update=True): 
    
        '''
        param grids: the 1st element is the number of grids for mean prediction, 
                    the 2nd is for var
        '''       
        
        self.grid_num = grid_num
        self.l_bound, self.u_bound = bounds
        self.cutoffs = GP.cutoffs
        self.bodies = bodies
        if not load_grid:
            y_mean, y_var = self.GenGrid(GP, bond_struc, update)
        else:
            y_mean = np.load('grid3_mean.npy')
            y_var = np.load('grid3_var.npy')
#            y_mean, y_var = utils.merge(load_grid, noa, nop)
        self.build_map(y_mean, y_var, svd_rank=svd_rank, load_svd=load_svd) 


    #@profile
    def GenGrid(self, GP, bond_struc, update, processes=1):#, processes=mp.cpu_count()):
    
        '''
        generate grid data of mean prediction and L^{-1}k* for each triplet
         implemented in a parallelized style
        '''
        # ------ change GP kernel to 3 body ------
        original_kernel = GP.kernel
        original_hyps = np.copy(GP.hyps)
        GP.kernel = three_body
        GP.hyps = [GP.hyps[-3], GP.hyps[-2], GP.hyps[-1]]

        # ------ construct grids ------
        nop = self.grid_num[0]
        noa = self.grid_num[2]
        bond_lengths = np.linspace(self.l_bound[0], self.u_bound[0], nop)
        angles = np.linspace(self.l_bound[2], self.u_bound[2], noa)
        bond_means = np.zeros([nop, nop, noa])
        bond_vars = np.zeros([nop, nop, noa, len(GP.alpha)])
        env12 = env.AtomicEnvironment(bond_struc, 0, self.cutoffs)

        pool_list = [(i, angles[i], bond_lengths, GP, env12, update)\
                     for i in range(noa)]
        pool = mp.Pool(processes=processes)
        if not update:
            subprocess.run(['rm', '-r', 'kv3'])
            subprocess.run(['mkdir', 'kv3'])
        A_list = pool.map(self._GenGrid_inner, pool_list)

        for a12 in range(noa):
            bond_means[:, :, a12] = A_list[a12][0]
            bond_vars[:, :, a12, :] = A_list[a12][1]
        pool.close()
        pool.join()

        # ------ change back to original GP ------
        GP.hyps = original_hyps
        GP.kernel = original_kernel
      
        # ------ save mean and var to file -------
        np.save('grid3_mean', bond_means)
        np.save('grid3_var', bond_vars)
        return bond_means, bond_vars

    def _GenGrid_inner(self, params):
    
        '''
        generate grid for each angle, used to parallelize grid generation
        '''
        a12, angle12, bond_lengths, GP, env12, update = params
        nop = self.grid_num[0]
        noa = self.grid_num[2]
        angle12 = angle12
        bond_means = np.zeros([nop, nop])
        bond_vars = np.zeros([nop, nop, len(GP.alpha)])

        # open saved k vector file, and write to new file
        kv_filename = 'kv3/'+str(a12)
        size = len(GP.training_data) * 3
        new_kv_file = np.zeros((nop**2+1, size))
        new_kv_file[0,0] = size

        if update:
            old_kv_file = np.load(kv_filename+'.npy') 
            last_size = int(old_kv_file[0,0])
            new_kv_file[:, :last_size] = old_kv_file
            ds = [1, 2, 3]

        for b1, r1 in enumerate(bond_lengths):
            r1 = bond_lengths[b1]
            for b2, r2 in enumerate(bond_lengths):
                x2 = r2 * np.cos(angle12)
                y2 = r2 * np.sin(angle12)
                r12 = np.linalg.norm(np.array([x2-r1, y2, 0]))

                env12.bond_array_3 = np.array([[r1, 1, 0, 0], [r2, 0, 0, 0]])
                env12.cross_bond_dists = np.array([[0, r12], [r12, 0]])

                if update:
                    # calculate kernel functions of those newly added training data
                    k12_v = new_kv_file[1+b1*nop+b2, :]
                    for m_index in range(last_size, size):
                        x_2 = GP.training_data[int(math.floor(m_index / 3))]
                        d_2 = ds[m_index % 3]
                        k12_v[m_index] = GP.kernel(env12, x_2, 1, d_2,
                                               GP.hyps, GP.cutoffs)
                else:
                    k12_v = GP.get_kernel_vector(env12, 1)   

                new_kv_file[1+b1*nop+b2, :] = k12_v

                # calculate mean and var value for the mapping
                v12_vec = solve_triangular(GP.l_mat, k12_v, lower=True)
                mean_diff = np.matmul(k12_v, GP.alpha)
                bond_means[b1, b2] = mean_diff
                bond_vars[b1, b2, :] = v12_vec  
                      
        # replace the old file with the new file
        np.save(kv_filename, new_kv_file)

        return bond_means, bond_vars


    def build_map(self, y_mean, y_var, svd_rank, load_svd):
    
        '''
        build 3-d spline function for mean, 
        3-d for the low rank approximation of L^{-1}k*
        '''
        nop = self.grid_num[0]
        noa = self.grid_num[2]
        self.mean = SplinesInterpolation(y_mean, u_bounds=self.u_bound, 
                    l_bounds=self.l_bound, orders=np.array([nop, nop, noa])) 

        self.var = PCASplines(y_var, u_bounds=self.u_bound, l_bounds=self.l_bound, 
                   orders=np.array([nop, nop, noa]), svd_rank=svd_rank, 
                   load_svd=load_svd)

    def build_selfkern(self, grid_kern):
        self.selfkern = TruncatedSVD(n_components=100, n_iter=7, random_state=42)
        self.selfkern.fit(grid_kern)
       
    def predict(self, atom_env, GP, mean_only):

        '''
        predict for an atom environment
        param: atom_env: ChemicalEnvironment
        return force on an atom with its variance
        '''
        t0 = time.time()
        bond_array = atom_env.bond_array_3
        cross_bond_inds = atom_env.cross_bond_inds
        cross_bond_dists = atom_env.cross_bond_dists
        triplets = atom_env.triplet_counts
        tri_12, tri_21, xyz_1s, xyz_2s = get_triplets(bond_array, 
            cross_bond_inds, cross_bond_dists, triplets)
        tri_12 = np.array(tri_12)
        tri_21 = np.array(tri_21)
        xyz_1s = np.array(xyz_1s)
        xyz_2s = np.array(xyz_2s)
        #print('\nget triplets', time.time()-t0)       

        # predict mean
        t0 = time.time()
        f0_12 = self.mean(tri_12)
        f0_21 = self.mean(tri_21)
        f12 = np.diag(f0_12) @ xyz_1s
        f21 = np.diag(f0_21) @ xyz_2s
        mgp_f = np.sum(f12 + f21, axis=0)
        #print('mean', time.time()-t0)

        # predict var        
        mgp_v = np.zeros(3)
        if not mean_only:
            t0 = time.time()
            self_kern = np.zeros(3)
            if self.bodies == '3':
                sig, ls, noise = GP.hyps
            elif self.bodies == '2+3':
                sig2, ls2, sig, ls, noise = GP.hyps
            r_cut = GP.cutoffs[1]
            for d in range(3):
                self_kern[d] = self_three_body_jit(bond_array,
                       cross_bond_inds, 
                       cross_bond_dists,
                       triplets, 
                       d+1, sig, ls, r_cut, quadratic_cutoff)
         #   print('self kern', time.time()-t0, ',value:', self_kern)

            t0 = time.time()
            v0_12 = self.var(tri_12)
            v0_21 = self.var(tri_21)
            v12 = v0_12 @ xyz_1s
            v21 = v0_21 @ xyz_2s
            v = v12 + v21
            if self.bodies == '3':
                mgp_v = - np.sum(v ** 2, axis=0) + self_kern
                return mgp_f, mgp_v
            elif self.bodies == '2+3':
                v = self.var.V @ v
                return mgp_f, self_kern, v
          #  print('var', time.time()-t0, ',value:', mgp_v)        
        else: 
            if self.bodies == '3':
                return mgp_f, mgp_v
            elif self.bodies == '2+3':
                return mgp_f, 0, 0

@njit
def get_triplets(bond_array, cross_bond_inds, 
                 cross_bond_dists, triplets):
    num = np.sum(triplets)
    tris1 = np.zeros((num,3))
    tris2 = np.zeros((num,3))
    tri_dir1 = np.zeros((num,3))
    tri_dir2 = np.zeros((num,3))

    k = 0
    for m in range(bond_array.shape[0]):
        r1 = bond_array[m, 0]
        c1 = bond_array[m, 1:]

        for n in range(triplets[m]):
            ind1 = cross_bond_inds[m, m+n+1]
            r2 = bond_array[ind1, 0]
            c2 = bond_array[ind1, 1:]
            a12 = np.arccos(np.sum(c1*c2))

            tris1[k] = np.array((r1, r2, a12))
            tris2[k] = np.array((r2, r1, a12))
            tri_dir1[k] = c1
            tri_dir2[k] = c2
            k += 1
    return tris1, tris2, tri_dir1, tri_dir2 


@njit
def quadratic_cutoff(r_cut, ri, ci):
    rdiff = r_cut - ri
    fi = rdiff * rdiff
    fdi = 2 * rdiff * ci

    return fi, fdi

@njit
def self_two_body_jit(bond_array, d, sig, ls,
                      r_cut, cutoff_func):
    kern = 0

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    sig2 = sig*sig

    for m in range(bond_array.shape[0]):
        ri = bond_array[m, 0]
        ci = bond_array[m, d]
        fi, fdi = cutoff_func(r_cut, ri, ci)

        for n in range(m, bond_array.shape[0]):
            rj = bond_array[n, 0]
            cj = bond_array[n, d]
            fj, fdj = cutoff_func(r_cut, rj, cj)
            r11 = ri - rj

            A = ci * cj
            B = r11 * ci
            C = r11 * cj
            D = r11 * r11

            if m == n:
                kern += force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2,
                                 ls3, sig2)
            else:
                kern += 2*force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2,
                                 ls3, sig2)
    return kern


@njit
def self_three_body_jit(bond_array, cross_bond_inds, 
                   cross_bond_dists, triplets,
                   d, sig, ls, r_cut, cutoff_func):
    kern = 0

    # pre-compute constants that appear in the inner loop
    sig2 = sig*sig
    ls1 = 1 / (2*ls*ls)
    ls2 = 1 / (ls*ls)
    ls3 = ls2*ls2
    
    for m in range(bond_array.shape[0]):
        ri1 = bond_array[m, 0]
        ci1 = bond_array[m, d]
        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)

        for n in range(triplets[m]):
            ind1 = cross_bond_inds[m, m+n+1]
            ri2 = bond_array[ind1, 0]
            ci2 = bond_array[ind1, d]
            fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)

            ri3 = cross_bond_dists[m, m+n+1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)

            fi = fi1*fi2*fi3
            fdi = fdi1*fi2*fi3+fi1*fdi2*fi3

            for p in range(m, bond_array.shape[0]):
                rj1 = bond_array[p, 0]
                cj1 = bond_array[p, d]
                fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)

                for q in range(triplets[p]):
                    ind2 = cross_bond_inds[p, p+1+q]
                    rj2 = bond_array[ind2, 0]
                    cj2 = bond_array[ind2, d]
                    fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)

                    rj3 = cross_bond_dists[p, p+1+q]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)

                    fj = fj1*fj2*fj3
                    fdj = fdj1*fj2*fj3+fj1*fdj2*fj3

                    tri_kern = triplet_kernel(ci1, ci2, cj1, cj2, ri1, ri2, ri3,
                                           rj1, rj2, rj3, fi, fj, fdi, fdj,
                                           ls1, ls2, ls3, sig2)
                    if p == m:
                        kern += tri_kern
                    else:
                        kern += 2 * tri_kern

    return kern


@njit
def triplet_kernel(ci1, ci2, cj1, cj2, ri1, ri2, ri3, rj1, rj2, rj3, fi, fj,
                   fdi, fdj, ls1, ls2, ls3, sig2):
    r11 = ri1-rj1
    r12 = ri1-rj2
    r13 = ri1-rj3
    r21 = ri2-rj1
    r22 = ri2-rj2
    r23 = ri2-rj3
    r31 = ri3-rj1
    r32 = ri3-rj2
    r33 = ri3-rj3

    # sum over all six permutations
    M1 = three_body_helper_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)
    M2 = three_body_helper_2(ci2, ci1, cj2, cj1, r21, r13, r32, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)
    M3 = three_body_helper_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)
    M4 = three_body_helper_1(ci1, ci2, cj2, cj1, r12, r21, r33, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)
    M5 = three_body_helper_2(ci2, ci1, cj1, cj2, r22, r13, r31, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)
    M6 = three_body_helper_2(ci1, ci2, cj2, cj1, r11, r23, r32, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, sig2)

    return M1 + M2 + M3 + M4 + M5 + M6


@njit
def three_body_helper_1(ci1, ci2, cj1, cj2, r11, r22, r33,
                        fi, fj, fdi, fdj,
                        ls1, ls2, ls3, sig2):
    A = ci1*cj1+ci2*cj2
    B = r11*ci1+r22*ci2
    C = r11*cj1+r22*cj2
    D = r11*r11+r22*r22+r33*r33

    M = force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2)

    return M


@njit
def three_body_helper_2(ci1, ci2, cj1, cj2, r12, r23, r31,
                        fi, fj, fdi, fdj,
                        ls1, ls2, ls3, sig2):
    A = ci1*cj2
    B = r12*ci1+r23*ci2
    C = r12*cj2+r31*cj1
    D = r12*r12+r23*r23+r31*r31

    M = force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2)

    return M


@njit
def force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2):
    E = exp(-D * ls1)
    F = B * fi
    G = -C * fj
    I = fdi * fdj
    J = F * fdj
    K = G * fdi
    L = A * fi * fj + F * G * ls2
    M = sig2 * (I + (J + K + L) * ls2) * E

    return M




import numpy as np
import datetime
import time

from ase.md import MDLogger
from ase import units

from flare.modules.gp_calculator import GPCalculator

class OTFLogger(MDLogger):

    def __init__(self, dyn, atoms, logfile, header=False, stress=False,
            peratom=False, mode="w"):

        super().__init__(dyn, atoms, logfile, header, stress,
                peratom, mode)

        self.write_header_info()
        self.start_time = time.time()


    def write_header_info(self):
        gp_model = self.atoms.calc.gp_model
        self.logfile.write(str(datetime.datetime.now()))
        self.logfile.write('\nnumber of cpu cores: ') # TODO
        self.logfile.write('\ncutoffs: '+str(gp_model.cutoffs))
        self.logfile.write('\nkernel: '+gp_model.kernel.__name__)
        self.logfile.write('\nnumber of parameters: '+str(len(gp_model.hyps)))
        self.logfile.write('\nhyperparameters: '+str(gp_model.hyps))
        self.logfile.write('\nhyperparameter optimization algorithm: '+gp_model.algo)
        self.logfile.write('\nuncertainty tolerance: '+str(self.dyn.std_tolerance))
        self.logfile.write('\nperiodic cell:\n'+str(self.atoms.cell))
        self.logfile.write('\n')


    def __call__(self):
        self.logfile.write('\n'+50*'-')
        if self.dyn is not None:
            steps = self.dyn.get_time()
            t = steps / (1000*units.fs)
            if type(self.atoms.calc) != GPCalculator: # TODO
                self.logfile.write('\n-*Frame: '+str(steps))
            else:
                self.logfile.write('\n-Frame: '+str(steps))
            self.logfile.write('\nSimulation time: '+str(t))
        self.logfile.write('\n')

        # add positions, forces and stds to be written
        positions = self.atoms.get_positions()
        forces = self.atoms.get_forces()
        velocities = self.atoms.get_velocities()
        if type(self.atoms.calc) == GPCalculator: # TODO: make it compatible with mff
            stds = self.atoms.get_uncertainties()
            force_str = 'GP Forces'
        else:
            stds = np.zeros(positions.shape)
            force_str = 'DFT Forces'
        self.logfile.write((10*' ')+'Positions'+(11*' '+'|'+11*' ')+
                                    force_str+(11*' '+'|'+11*' ')+
                                    'Velocities'+(10*' '+'|'+11*' ')+
                                    'Uncertainties\n')

        template = '{:9f} {:9f} {:9f} | {:9f} {:9f} {:9f} | {:9f} {:9f} {:9f} | {:9f} {:9f} {:9f}'
        for atom in range(len(positions)):
            dat = template.format(\
                    positions[atom][0], positions[atom][1], positions[atom][2],
                    forces[atom][0],    forces[atom][1],    forces[atom][2],
                    velocities[atom][0],velocities[atom][1],velocities[atom][2],
                    stds[atom][0],      stds[atom][1],      stds[atom][2])
            self.logfile.write(dat+'\n')

        # write energy, temperature info
        epot = self.atoms.get_potential_energy()
        ekin = self.atoms.get_kinetic_energy()
        temp = ekin / (1.5 * units.kB * self.natoms)
        if self.peratom:
            epot /= self.natoms
            ekin /= self.natoms
        self.logfile.write('\ntotal energy: '+str(epot+ekin))
        self.logfile.write('\nkinetic energy: '+str(ekin))
        self.logfile.write('\ntemperature: '+str(temp))
        self.logfile.write('\nwall time from start: '+str(time.time()-self.start_time))

        self.logfile.write('\n')
        self.logfile.flush()
        

    def add_atom_info(self, target_atom, uncertainty):
        self.logfile.write('\nAdding atom '+str(target_atom)+' to the training set with uncertainty: '+str(uncertainty))



"""
any DFT interface can be added here, as long as two functions listed below are implemented

* parse_dft_input(dft_input)
* dft_module.run_dft_par(dft_input, structure, dft_loc, no_cpus)
"""

from flare.dft_interface import qe_util, cp2k_util
dft_software = { "qe":qe_util,
                 "cp2k":cp2k_util}



"""The :class:`Structure` object is a collection of atoms in a periodic box. The mandatory inputs are the cell vectors of the box and the chemical species and Cartesian coordinates of the atoms. The atoms are automatically folded back into the primary cell, so the input coordinates don't need to lie inside the box. Energy, force, and stress information can be provided for training machine learned force fields."""
import numpy as np
from typing import List
from flare.util import element_to_Z, NumpyEncoder
from json import dumps

try:
    # Used for to_pmg_structure method
    import pymatgen.core.structure as pmgstruc
    _pmg_present = True
except ImportError:
    _pmg_present = False

class Structure:
    """
    Contains information about a structure of atoms, including the periodic
    cell boundaries and atomic species and coordinates.

    :param cell: 3x3 array whose rows are the Bravais lattice vectors of the \
cell.
    :type cell: np.ndarray
    :param species: List of atomic species, which are represented either as \
integers or chemical symbols.
    :type species: List
    :param positions: Nx3 array of atomic coordinates.
    :type positions: np.ndarray
    :param mass_dict: Dictionary of atomic masses used in MD simulations.
    :type mass_dict: dict
    :param prev_positions: Nx3 array of previous atomic coordinates used in \
MD simulations.
    :type prev_positions: np.ndarray
    :param species_labels: List of chemical symbols. Used in the output file \
of on-the-fly runs.
    :type species_labels: List[str]
    """

    def __init__(self, cell, species, positions, mass_dict=None,
                 prev_positions=None, species_labels=None):
        self.cell = cell
        self.vec1 = cell[0, :]
        self.vec2 = cell[1, :]
        self.vec3 = cell[2, :]

        # get cell matrices for wrapping coordinates
        self.cell_transpose = self.cell.transpose()
        self.cell_transpose_inverse = np.linalg.inv(self.cell_transpose)
        self.cell_dot = self.get_cell_dot()
        self.cell_dot_inverse = np.linalg.inv(self.cell_dot)

        # set positions
        self.positions = np.array(positions)
        self.wrap_positions()

        # If species are strings, convert species to integers by atomic number
        if species_labels is None:
            self.species_labels = species
        else:
            self.species_labels = species_labels
        self.coded_species = np.array([element_to_Z(spec) for spec in species])
        self.nat = len(species)

        # Default: atoms have no velocity
        if prev_positions is None:
            self.prev_positions = np.copy(self.positions)
        else:
            assert len(positions) == len(prev_positions), 'Previous ' \
                                                          'positions and ' \
                                                          'positions are not' \
                                                          'same length'
            self.prev_positions = prev_positions

        self.energy = None
        self.stress = None
        self.forces = np.zeros((len(positions), 3))
        self.stds = np.zeros((len(positions), 3))
        self.mass_dict = mass_dict

    def get_cell_dot(self):
        """
        Compute 3x3 array of dot products of cell vectors used to fold atoms \
back to the unit cell.

        :return: 3x3 array of cell vector dot products.
        :rtype: np.ndarray
        """

        cell_dot = np.zeros((3, 3))

        for m in range(3):
            for n in range(3):
                cell_dot[m, n] = np.dot(self.cell[m], self.cell[n])

        return cell_dot

    @staticmethod
    def raw_to_relative(positions, cell_transpose, cell_dot_inverse):
        """Convert Cartesian coordinates to relative coordinates expressed \
in terms of the cell vectors.

        :param positions: Cartesian coordinates.
        :type positions: np.ndarray
        :param cell_transpose: Transpose of the cell array.
        :type cell_transpose: np.ndarray
        :param cell_dot_inverse: Inverse of the array of dot products of \
cell vectors.
        :type cell_dot_inverse: np.ndarray
        :return: Relative positions.
        :rtype: np.ndarray
        """

        relative_positions = \
            np.matmul(np.matmul(positions, cell_transpose),
                      cell_dot_inverse)

        return relative_positions

    @staticmethod
    def relative_to_raw(relative_positions, cell_transpose_inverse,
                        cell_dot):
        positions = \
            np.matmul(np.matmul(relative_positions, cell_dot),
                      cell_transpose_inverse)

        return positions

    def wrap_positions(self):
        rel_pos = \
            self.raw_to_relative(self.positions, self.cell_transpose,
                                 self.cell_dot_inverse)

        rel_wrap = rel_pos - np.floor(rel_pos)

        pos_wrap = self.relative_to_raw(rel_wrap, self.cell_transpose_inverse,
                                        self.cell_dot)

        self.wrapped_positions = pos_wrap

    def indices_of_specie(self, specie: int):
        """
        Return the indicies of a given species.

        :param specie:
        :return:
        """
        return [i for i, spec in enumerate(self.coded_species)
                if spec == specie]

    # TODO make more descriptive
    def __str__(self):
        return 'Structure with {} atoms of types {}'.format(self.nat,
                                                     set(self.species_labels))

    def __len__(self):
        return self.nat

    def as_dict(self):
        """
        Returns structure as a dictionary for serialization purposes.

        :return:
        """
        return dict(vars(self))

    def as_str(self):
        return dumps(self.as_dict(), cls=NumpyEncoder)

    @staticmethod
    def from_dict(dictionary):
        struc = Structure(cell=np.array(dictionary['cell']),
                          positions=np.array(dictionary['positions']),
                          species=dictionary['coded_species'])

        struc.forces = np.array(dictionary['forces'])
        struc.stress = dictionary['stress']
        struc.stds = np.array(dictionary['stds'])
        struc.mass_dict = dictionary['mass_dict']
        struc.species_labels = dictionary['species_labels']

        return struc

    def to_pmg_structure(self):
        """
        Returns FLARE structure as a pymatgen structure.
        :return: Pymatgen structure corresponding to current FLARE structure
        """

        if not _pmg_present:
            raise ModuleNotFoundError("Pymatgen is not present. Please "
                                      "install Pymatgen and try again")

        site_properties = {'force:': self.forces, 'std': self.stds}

        return pmgstruc.Structure(lattice=self.cell,
                                  species=self.species_labels,
                                  coords=self.positions,
                                  coords_are_cartesian=True,
                                  site_properties=site_properties
                                  )

    @staticmethod
    def from_pmg_structure(structure):
        """
        Returns Pymatgen structure as FLARE structure.
        
        :param structure: Pymatgen structure
        :return: FLARE Structure
        """

        cell = structure.lattice.matrix
        species = [str(spec) for spec in structure.species]
        positions = structure.cart_coords

        new_struc = Structure(cell=cell,species=species,
                              positions=positions)

        site_props = structure.site_properties

        if 'force' in site_props.keys():
            forces = site_props['force']
            new_struc.forces = [np.array(force) for force in forces]

        if 'std' in site_props.keys():
            stds = site_props['std']
            new_struc.stds = [np.array(std) for std in stds]

        return new_struc

def get_unique_species(species):
    unique_species = []
    coded_species = []
    for spec in species:
        if spec in unique_species:
            coded_species.append(unique_species.index(spec))
        else:
            coded_species.append(len(unique_species))
            unique_species.append(spec)
    coded_species = np.array(coded_species)

    return unique_species, coded_species


if __name__ == '__main__':
    pass



from math import cos, sin, pi
from numba import njit


@njit
def hard_cutoff(r_cut, ri, ci):
    return 1, 0


@njit
def quadratic_cutoff(r_cut, ri, ci):
    rdiff = r_cut - ri
    fi = rdiff * rdiff
    fdi = 2 * rdiff * ci

    return fi, fdi


@njit
def cubic_cutoff(r_cut, ri, ci):
    rdiff = r_cut - ri
    fi = rdiff * rdiff * rdiff
    fdi = 3 * rdiff * rdiff * ci

    return fi, fdi


@njit
def cosine_cutoff(r_cut, ri, ci, d=1):
    if ri > r_cut - d:
        fi = (1/2) * (cos(pi * (ri - r_cut + d) / d) + 1)
        fdi = (pi/(2 * d)) * sin(pi * (r_cut - ri) / d) * ci
    else:
        fi = 1
        fdi = 0

    return fi, fdi



import os
from subprocess import call
import time
import numpy as np
from flare import struc
from typing import List


def run_dft(qe_input, structure, dft_loc):
    run_qe_path = qe_input
    edit_dft_input_positions(run_qe_path, structure)
    qe_command = '{0} < {1} > {2}'.format(dft_loc, run_qe_path,
                                                 'pwscf.out')
    call(qe_command, shell=True)

    return parse_dft_forces('pwscf.out')


def run_dft_par(qe_input, structure, dft_loc, no_cpus):
    run_qe_path = qe_input
    edit_dft_input_positions(run_qe_path, structure)
    qe_command = \
        'mpirun -np {0} {1} < {2} > {3}'.format(no_cpus, dft_loc, run_qe_path,
                                                'pwscf.out')

    call(qe_command, shell=True)

    return parse_dft_forces('pwscf.out')


def run_dft_en_par(qe_input, structure, dft_loc, no_cpus):
    run_qe_path = qe_input
    edit_dft_input_positions(run_qe_path, structure)
    qe_command = \
        'mpirun -np {0} {1} < {2} > {3}'.format(no_cpus, dft_loc, run_qe_path,
                                                'pwscf.out')
    call(qe_command, shell=True)

    forces, energy = parse_dft_forces_and_energy('pwscf.out')

    return forces, energy


def run_dft_npool(qe_input, qe_output, structure, dft_loc, npool):
    run_qe_path = qe_input
    edit_dft_input_positions(run_qe_path, structure)
    qe_command = \
        'mpirun {0} -npool {1} < {2} > {3}'.format(dft_loc, npool, run_qe_path,
                                                   qe_output)
    call(qe_command, shell=True)

    return parse_dft_forces(qe_output)


def run_dft_command(qe_input, qe_output, dft_loc, npool):
    qe_command = \
        'mpirun {0} -npool {1} < {2} > {3}'.format(dft_loc, npool, qe_input,
                                                   qe_output)
    call(qe_command, shell=True)

    return parse_dft_forces(qe_output)


def parse_dft_input(qe_input: str):
    positions = []
    species = []
    cell = []

    with open(qe_input) as f:
        lines = f.readlines()

    # Find the cell and positions in the output file
    cell_index = None
    positions_index = None
    nat = None
    species_index = None

    for i, line in enumerate(lines):
        if 'CELL_PARAMETERS' in line:
            cell_index = int(i + 1)
        if 'ATOMIC_POSITIONS' in line:
            positions_index = int(i + 1)
        if 'nat' in line:
            nat = int(line.split('=')[1])
        if 'ATOMIC_SPECIES' in line:
            species_index = int(i + 1)

    assert cell_index is not None, 'Failed to find cell in input'
    assert positions_index is not None, 'Failed to find positions in input'
    assert nat is not None, 'Failed to find number of atoms in input'
    assert species_index is not None, 'Failed to find atomic species in input'

    # Load cell
    for i in range(cell_index, cell_index + 3):
        cell_line = lines[i].strip()
        cell.append(np.fromstring(cell_line, sep=' '))
    cell = np.array(cell)

    # Check cell IO
    assert len(cell) != 0, 'Cell failed to load'
    assert np.shape(cell) == (3, 3), 'Cell failed to load correctly'

    # Load positions
    for i in range(positions_index, positions_index + nat):
        line_string = lines[i].strip().split()
        species.append(line_string[0])

        pos_string = ' '.join(line_string[1:4])

        positions.append(np.fromstring(pos_string, sep=' '))
    # Check position IO
    assert positions != [], "Positions failed to load"
    positions = np.array(positions)

    # see conversions.nb for conversion from amu to md units
    massconvert = 0.000103642695727
    masses = {}
    for i in range(species_index, species_index + len(set(species))):
        # Expects lines of format like: H 1.0 H_pseudo_name.ext
        line = lines[i].strip().split()
        masses[line[0]] = float(line[1]) * massconvert

    return positions, species, cell, masses


def dft_input_to_structure(qe_input: str):
    """
    Parses a qe input and returns the atoms in the file as a Structure object
    :param qe_input: QE Input file to parse
    :return:
    """
    positions, species, cell, masses = parse_dft_input(qe_input)
    _, coded_species = struc.get_unique_species(species)
    return struc.Structure(positions=positions, species=coded_species,
                           cell=cell, mass_dict=masses, species_labels=species)


def edit_dft_input_positions(qe_input: str, structure):
    """
    Write the current configuration of the OTF structure to the
    qe input file
    """

    with open(qe_input, 'r') as f:
        lines = f.readlines()

    file_pos_index = None
    cell_index = None
    nat = None
    for i, line in enumerate(lines):
        if 'ATOMIC_POSITIONS' in line:
            file_pos_index = int(i + 1)
        if 'CELL_PARAMETERS' in line:
            cell_index = int(i + 1)
        # Load nat into variable then overwrite it with new nat
        if 'nat' in line:
            nat = int(line.split('=')[1])
            nat_index = int(i)
            lines[nat_index] = 'nat = ' + str(structure.nat) + '\n'

    assert file_pos_index is not None, 'Failed to find positions in input'
    assert cell_index is not None, 'Failed to find cell in input'
    assert nat is not None, 'Failed to find nat in input'

    # TODO Catch case where the punchout structure has more atoms than the
    # original structure

    for pos_index, line_index in enumerate(
            range(file_pos_index, file_pos_index + structure.nat)):
        pos_string = ' '.join([structure.species_labels[pos_index],
                               str(structure.positions[pos_index][
                                       0]),
                               str(structure.positions[pos_index][
                                       1]),
                               str(structure.positions[pos_index][
                                       2])])
        if line_index < len(lines):
            lines[line_index] = str(pos_string + '\n')
        else:
            lines.append(str(pos_string + '\n'))

    # TODO current assumption: if there is a new structure, then the new
    # structure has fewer atoms than the  previous one. If we are always
    # 'editing' a version of the larger structure than this will be okay with
    # the punchout method.
    for line_index in range(file_pos_index + structure.nat,
                            file_pos_index + nat):
        lines[line_index] = ''

    lines[cell_index] = ' '.join([str(x) for x in structure.vec1]) + '\n'
    lines[cell_index + 1] = ' '.join([str(x) for x in structure.vec2]) \
                            + '\n'
    lines[cell_index + 2] = ' '.join([str(x) for x in structure.vec3]) \
                            + '\n'

    with open(qe_input, 'w') as f:
        for line in lines:
            f.write(line)


def parse_dft_forces(outfile: str):
    """
    Get forces from a pwscf file in eV/A

    :param outfile: str, Path to pwscf output file
    :return: list[nparray] , List of forces acting on atoms
    """
    forces = []
    total_energy = np.nan

    with open(outfile, 'r') as outf:
        for line in outf:
            if line.lower().startswith('!    total energy'):
                total_energy = float(line.split()[-2])

            if line.find('force') != -1 and line.find('atom') != -1:
                line = line.split('force =')[-1]
                line = line.strip()
                line = line.split(' ')
                line = [x for x in line if x != '']
                temp_forces = []
                for x in line:
                    temp_forces.append(float(x))
                forces.append(np.array(list(temp_forces)))

    assert total_energy != np.nan, "Quantum ESPRESSO parser failed to read " \
                                   "the file {}. Run failed.".format(outfile)

    # Convert from ry/au to ev/angstrom
    conversion_factor = 25.71104309541616

    forces = [conversion_factor * force for force in forces]
    forces = np.array(forces)

    return forces


def parse_dft_forces_and_energy(outfile: str):
    """
    Get forces from a pwscf file in eV/A

    :param outfile: str, Path to pwscf output file
    :return: list[nparray] , List of forces acting on atoms
    """
    forces = []
    total_energy = np.nan

    with open(outfile, 'r') as outf:
        for line in outf:
            if line.lower().startswith('!    total energy'):
                total_energy = float(line.split()[-2])

            if line.find('force') != -1 and line.find('atom') != -1:
                line = line.split('force =')[-1]
                line = line.strip()
                line = line.split(' ')
                line = [x for x in line if x != '']
                temp_forces = []
                for x in line:
                    temp_forces.append(float(x))
                forces.append(np.array(list(temp_forces)))

    assert total_energy != np.nan, "Quantum ESPRESSO parser failed to read " \
                                   "the file {}. Run failed.".format(outfile)

    # Convert from ry/au to ev/angstrom
    conversion_factor = 25.71104309541616

    forces = [conversion_factor * force for force in forces]
    forces = np.array(forces)

    return forces, total_energy



import numpy as np
from numpy import array
from numba import njit
import io, os, sys, time, random, logging
import multiprocessing as mp
import concurrent.futures
import cProfile

import flare.gp as gp
import flare.env as env
import flare.struc as struc
import flare.kernels as kernels
from flare.env import AtomicEnvironment
from flare.kernels import triplet_kernel, three_body_helper_1, \
    three_body_helper_2, force_helper
from flare.cutoffs import quadratic_cutoff


def save_GP(GP, prefix):
    np.save(prefix+'alpha', GP.alpha)
    np.save(prefix+'hyps', GP.hyps)
    np.save(prefix+'l_mat', GP.l_mat)


def load_GP(GP, prefix):
    GP.alpha = np.load(prefix+'alpha.npy')
    GP.hyps = np.load(prefix+'hyps.npy')
    GP.l_mat = np.load(prefix+'l_mat.npy')
    l_mat_inv = np.linalg.inv(GP.l_mat)
    GP.ky_mat_inv = l_mat_inv.T @ l_mat_inv


def save_grid(bond_lens, bond_ens_diff, bond_vars_diff, prefix):
    np.save(prefix+'-bond_lens', bond_lens)
    np.save(prefix+'-bond_ens_diff', bond_ens_diff)
    np.save(prefix+'-bond_vars_diff', bond_vars_diff) 


def load_grid(prefix):
    bond_lens = np.load(prefix+'bond_lens.npy')
    bond_ens_diff = np.load(prefix+'bond_ens_diff.npy')
    bond_vars_diff = np.load(prefix+'bond_vars_diff.npy')  
    return bond_lens, bond_ens_diff, bond_vars_diff


def merge(prefix, a_num, g_num):
    grid_means = np.zeros((g_num, g_num, a_num))
    grid_vars = np.zeros((g_num, g_num, a_num, g_num, g_num, a_num))
    for a12 in range(a_num):
        grid_means[:,:,a12] = np.load(prefix+str((a12, 0))+'-bond_means.npy')
        for a34 in range(a_num):
            grid_vars[:,:,a12,:,:,a34] = np.load(prefix+str((a12, a34))+'-bond_vars.npy')
    return grid_means, grid_vars


def svd_grid(matr, rank):
    u, s, vh = np.linalg.svd(matr, full_matrices=False)
    return u[:,:rank], s[:rank], vh[:rank, :]


def get_l_bound(curr_l_bound, structure, two_d=False):
    positions = structure.positions
    if two_d:
        cell = structure.cell[:2]
    else:
        cell = structure.cell

    min_dist = curr_l_bound
    for ind1, pos01 in enumerate(positions):
        for i1 in range(2):
            for vec1 in cell:
                pos1 = pos01 + i1 * vec1

                for ind2, pos02 in enumerate(positions):
                    for i2 in range(2):
                        for vec2 in cell:
                            pos2 = pos02 + i2 * vec2

                            if np.all(pos1 == pos2):
                                continue
                            dist12 = np.linalg.norm(pos1-pos2)
                            if dist12 < min_dist:
                                min_dist = dist12
                                min_atoms = (ind1, ind2)
    return min_dist


@njit
def get_bonds(ctype, etypes, bond_array):
    exist_species = []
    bond_lengths = []
    bond_dirs = []
    for i in range(len(bond_array)):
        bond = bond_array[i]
        spc = sorted([ctype, etypes[i]])
        if spc in exist_species:
            ind = exist_species.index(spc)
            bond_lengths[ind].append([bond[0]])
            bond_dirs[ind].append(bond[1:])
        else:
            exist_species.append(spc)
            bond_lengths.append([[bond[0]]])
            bond_dirs.append([bond[1:]])
    return exist_species, bond_lengths, bond_dirs


#@njit
def add_triplets(spcs_list, exist_species, tris, tri_dir, 
        r1, r2, a12, c1, c2):
    for i in range(2):
        spcs = spcs_list[i]
        if spcs not in exist_species:
            exist_species.append(spcs)
            tris.append([])
            tri_dir.append([])

        k = exist_species.index(spcs)
        triplet = (r2, r1, a12) if i else (r1, r2, a12)
        coord = c2 if i else c1
        tris[k].append(triplet)
        tri_dir[k].append(coord)
    return exist_species, tris, tri_dir


@njit
def get_triplets(ctype, etypes, bond_array, cross_bond_inds,  
                 cross_bond_dists, triplets):
    exist_species = []
    tris = []
    tri_dir = []

    for m in range(bond_array.shape[0]):
        r1 = bond_array[m, 0]
        c1 = bond_array[m, 1:]
        spc1 = etypes[m]

        for n in range(triplets[m]):
            ind1 = cross_bond_inds[m, m+n+1]
            r2 = bond_array[ind1, 0]
            c2 = bond_array[ind1, 1:]
            a12 = np.arccos(np.sum(c1*c2))
            spc2 = etypes[ind1]

#            if spc1 == spc2:
#                spcs_list = [[ctype, spc1, spc2], [ctype, spc1, spc2]]
#            elif ctype == spc1: # spc1 != spc2
#                spcs_list = [[ctype, spc1, spc2], [spc2, ctype, spc1]]
#            elif ctype == spc2: # spc1 != spc2
#                spcs_list = [[spc1, spc2, ctype], [spc2, ctype, spc1]]
#            else: # all different
#                spcs_list = [[ctype, spc1, spc2], [ctype, spc2, spc1]]
   
            spcs_list = [[ctype, spc1, spc2], [ctype, spc2, spc1]]
            for i in range(2):
                spcs = spcs_list[i]
                triplet = array([r2, r1, a12]) if i else array([r1, r2, a12]) 
                coord = c2 if i else c1
                if spcs not in exist_species:
                    exist_species.append(spcs)
                    tris.append([triplet])
                    tri_dir.append([coord])
                else:
                    k = exist_species.index(spcs)
                    tris[k].append(triplet)
                    tri_dir[k].append(coord)

    return exist_species, tris, tri_dir 


@njit
def self_two_body_mc_jit(bond_array, c, etypes,
                    d, sig, ls, r_cut, cutoff_func):

    """Multicomponent two-body force/force kernel accelerated with Numba's
    njit decorator.

    Loops over bonds in two environments and adds to the kernel if bonds are
    of the same type.
    """

    kern = 0

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    sig2 = sig * sig

    for m in range(bond_array.shape[0]):
        ri = bond_array[m, 0]
        ci = bond_array[m, d]
        fi, fdi = cutoff_func(r_cut, ri, ci)
        e1 = etypes[m]

        for n in range(m, bond_array.shape[0]):
            e2 = etypes[n]

            # check if bonds agree
            if (c == c and e1 == e2) or (c == e2 and c == e1):
                rj = bond_array[n, 0]
                cj = bond_array[n, d]
                fj, fdj = cutoff_func(r_cut, rj, cj)
                r11 = ri - rj

                A = ci * cj
                B = r11 * ci
                C = r11 * cj
                D = r11 * r11

                kern0 = force_helper(A, B, C, D, fi, fj, fdi, fdj,
                                     ls1, ls2, ls3, sig2)
                kern += kern0 if m == n else 2 * kern0

    return kern


@njit
def self_three_body_mc_jit(bond_array, cross_bond_inds, cross_bond_dists,
                 triplets, c, etypes, d, sig, ls, r_cut, cutoff_func):
    kern = 0

    # pre-compute constants that appear in the inner loop
    sig2 = sig*sig
    ls1 = 1 / (2*ls*ls)
    ls2 = 1 / (ls*ls)
    ls3 = ls2*ls2

    for m in range(bond_array.shape[0]):
        ri1 = bond_array[m, 0]
        ci1 = bond_array[m, d]
        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
        ei1 = etypes[m]

        for n in range(triplets[m]):
            ind1 = cross_bond_inds[m, m+n+1]
            ri2 = bond_array[ind1, 0]
            ci2 = bond_array[ind1, d]
            fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
            ei2 = etypes[ind1]

            ri3 = cross_bond_dists[m, m+n+1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)

            fi = fi1*fi2*fi3
            fdi = fdi1*fi2*fi3+fi1*fdi2*fi3

            for p in range(m, bond_array.shape[0]):
                rj1 = bond_array[p, 0]
                cj1 = bond_array[p, d]
                fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)
                ej1 = etypes[p]

                for q in range(triplets[p]):
                    ind2 = cross_bond_inds[p, p+1+q]
                    rj2 = bond_array[ind2, 0]
                    cj2 = bond_array[ind2, d]
                    fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)
                    ej2 = etypes[ind2]

                    rj3 = cross_bond_dists[p, p+1+q]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)

                    fj = fj1*fj2*fj3
                    fdj = fdj1*fj2*fj3+fj1*fdj2*fj3

                    r11 = ri1-rj1
                    r12 = ri1-rj2
                    r13 = ri1-rj3
                    r21 = ri2-rj1
                    r22 = ri2-rj2
                    r23 = ri2-rj3
                    r31 = ri3-rj1
                    r32 = ri3-rj2
                    r33 = ri3-rj3

                    if (c == c):
                        if (ei1 == ej1) and (ei2 == ej2):
                            kern0 = \
                                three_body_helper_1(ci1, ci2, cj1, cj2, r11,
                                                    r22, r33, fi, fj, fdi, fdj,
                                                    ls1, ls2, ls3, sig2)
                        if (ei1 == ej2) and (ei2 == ej1):
                            kern0 = \
                                three_body_helper_1(ci1, ci2, cj2, cj1, r12,
                                                    r21, r33, fi, fj, fdi, fdj,
                                                    ls1, ls2, ls3, sig2)
                    if (c == ej1):
                        if (ei1 == ej2) and (ei2 == c):
                            kern0 = \
                                three_body_helper_2(ci2, ci1, cj2, cj1, r21,
                                                    r13, r32, fi, fj, fdi,
                                                    fdj, ls1, ls2, ls3, sig2)
                        if (ei1 == c) and (ei2 == ej2):
                            kern0 = \
                                three_body_helper_2(ci1, ci2, cj2, cj1, r11,
                                                    r23, r32, fi, fj, fdi,
                                                    fdj, ls1, ls2, ls3, sig2)
                    if (c == ej2):
                        if (ei1 == ej1) and (ei2 == c):
                            kern0 = \
                                three_body_helper_2(ci2, ci1, cj1, cj2, r22,
                                                    r13, r31, fi, fj, fdi,
                                                    fdj, ls1, ls2, ls3, sig2)
                        if (ei1 == c) and (ei2 == ej1):
                            kern0 = \
                                three_body_helper_2(ci1, ci2, cj1, cj2, r12,
                                                    r23, r31, fi, fj, fdi,
                                                    fdj, ls1, ls2, ls3, sig2)
                    kern += kern0 if m == p else 2 * kern0

    return kern



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    dependencies = fh.readlines()

setuptools.setup(
    name="mir-flare",
    packages=setuptools.find_packages(exclude=["tests"]),
    version="0.0.3",
    author="Materials Intelligence Research",
    author_email="mir@g.harvard.edu",
    description="Fast Learning of Atomistic Rare Events",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mir-group/flare",
    python_requires=">=3.6",
    install_requires=dependencies,
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Development Status :: 4 - Beta",
    ],
)



from flare.gp import GaussianProcess
from flare.struc import Structure
from flare.env import AtomicEnvironment
import numpy as np
import time
import datetime
import concurrent.futures
from flare import md
from flare.output import Output


class MD:
    """Generates NVE dynamics from a GP model."""

    def __init__(self, dt: float, number_of_steps: int, gp: GaussianProcess,
                 pos_init: np.ndarray, species, cell, masses,
                 prev_pos_init: np.ndarray=None, par: bool=False, skip: int=0,
                 output_name='otf_run.out'):

        self.dt = dt
        self.Nsteps = number_of_steps
        self.gp = gp

        self.structure = Structure(cell=cell, species=species,
                                   positions=pos_init,
                                   mass_dict=masses,
                                   prev_positions=prev_pos_init)

        self.noa = self.structure.positions.shape[0]
        self.atom_list = list(range(self.noa))
        self.curr_step = 0

        # choose prediction function
        if par is True:
            self.pred_func = self.predict_on_structure_par_en
        else:
            self.pred_func = self.predict_on_structure_en

        # initialize local energies
        self.local_energies = np.zeros(self.noa)

        self.pes = []
        self.kes = []

        self.output = Output(output_name)

    def run(self):
        self.output.write_header(self.gp.cutoffs, self.gp.kernel_name, self.gp.hyps,
                                 self.gp.algo, self.dt, self.Nsteps, self.structure)
        self.start_time = time.time()

        while self.curr_step < self.Nsteps:
            # verlet algorithm follows Frenkel p. 70
            self.pred_func()
            new_pos = md.update_positions(self.dt, self.noa, self.structure)
            self.update_temperature(new_pos)
            self.record_state()
            self.update_positions(new_pos)
            self.curr_step += 1

        self.output.conclude_run()

    def predict_on_structure_en(self):
        for n in range(self.structure.nat):
            chemenv = AtomicEnvironment(self.structure, n, self.gp.cutoffs)
            for i in range(3):
                force, var = self.gp.predict(chemenv, i + 1)
                self.structure.forces[n][i] = float(force)
                self.structure.stds[n][i] = np.sqrt(np.absolute(var))
            self.local_energies[n] = self.gp.predict_local_energy(chemenv)

    def predict_on_structure_par_en(self):
        n = 0
        with concurrent.futures.ProcessPoolExecutor() as executor:
            for res in executor.map(self.predict_on_atom_en, self.atom_list):
                for i in range(3):
                    self.structure.forces[n][i] = res[0][i]
                    self.structure.stds[n][i] = res[1][i]
                self.local_energies[n] = res[2]
                n += 1
        self.structure.dft_forces = False

    def predict_on_atom_en(self, atom):
        chemenv = AtomicEnvironment(self.structure, atom, self.gp.cutoffs)
        comps = []
        stds = []
        # predict force components and standard deviations
        for i in range(3):
            force, var = self.gp.predict(chemenv, i+1)
            comps.append(float(force))
            stds.append(np.sqrt(np.absolute(var)))

        # predict local energy
        local_energy = self.gp.predict_local_energy(chemenv)
        return comps, stds, local_energy

    def update_positions(self, new_pos):
        self.structure.prev_positions = self.structure.positions
        self.structure.positions = new_pos
        self.structure.wrap_positions()

    def update_temperature(self, new_pos):
        KE, temperature = \
                md.calculate_temperature(new_pos, self.structure, self.dt,
                                         self.noa)
        self.KE = KE
        self.temperature = temperature

    def record_state(self):
        self.pes.append(np.sum(self.local_energies))
        self.kes.append(self.KE)
        self.output.write_md_config(self.dt, self.curr_step, self.structure,
                                    self.temperature, self.KE, self.local_energies,
                                    self.start_time)
        self.output.write_xyz_config(self.curr_step, self.structure,
                                     self.dft_step)



import os
from subprocess import call
import time
import numpy as np
from flare import output
from flare import struc
from typing import List

name="CP2K"


def run_dft_par(dft_input, structure, dft_loc, no_cpus=1, dft_out="dft.out"):
    newfilename = edit_dft_input_positions(dft_input, structure)
    dft_command = \
        '{} -i {} > {}'.format(dft_loc, newfilename, dft_out)
    if (no_cpus > 1):
        dft_command = 'mpirun -np {} {}'.format(no_cpus, dft_command)
    # output.write_to_output(dft_command+'\n')
    call(dft_command, shell=True)
    os.remove(newfilename)

    return parse_dft_forces(dft_out)


def run_dft_en_par(dft_input, structure, dft_loc, no_cpus, dft_out="dft.out"):

    newfilename = edit_dft_input_positions(dft_input, structure)
    dft_command = \
        '{} -i {} > {}'.format(dft_loc, newfilename, dft_out)
    if (no_cpus > 1):
        dft_command = 'mpirun -np {} {}'.format(no_cpus, dft_command)
    # output.write_to_output(dft_command+'\n')
    call(dft_command, shell=True)
    os.remove(newfilename)

    forces, energy = parse_dft_forces_and_energy(dft_out)

    return forces, energy




def parse_dft_input(dft_input: str):
    positions = []
    species = []
    cell = []

    with open(dft_input) as f:
        lines = f.readlines()

    # Find the cell and positions in the output file
    cell_index = None
    positions_index = None
    nat = None
    # species_index = None

    for i, line in enumerate(lines):
        if '&CELL' in line:
            cell_index = int(i + 1)
        elif 'COORD' in line and 'END' not in line:
            positions_index = int(i + 1)
        elif ('&END' in line and (positions_index is not None) and (nat is None)):
            nat = i - positions_index
    #     if 'ATOMIC_SPECIES' in line:
    #         species_index = int(i + 1)

    assert cell_index is not None, 'Failed to find cell in input'
    assert positions_index is not None, 'Failed to find positions in input'
    assert nat is not None, 'Failed to find number of atoms in input'
    # assert species_index is not None, 'Failed to find atomic species in input'

    # Load cell
    # TO DO: allow to mess up the order of A, B, and C
    for i in range(cell_index, cell_index + 3):
        cell_line = list(map(float,lines[i].split()[1:]))
        cell.append(cell_line) #np.fromstring(cell_line[1:], sep=' '))
    cell = np.array(cell)

    # Check cell IO
    assert len(cell) != 0, 'Cell failed to load'
    assert np.shape(cell) == (3, 3), 'Cell failed to load correctly'

    # Load positions
    for i in range(positions_index, positions_index + nat):
        pos_line = lines[i].split()
        species.append(pos_line[0])
        # positions.append(np.fromstring(pos_string, sep=' '))
        positions.append(list(map(float, pos_line[1:])))

    # Check position IO
    assert positions != [], "Positions failed to load"
    positions = np.array(positions)

    # see conversions.nb for conversion from amu to md units
    ele_mass = {"H":1.007900, "He":4.002600, "Li":6.941000, "Be":9.012200, "B":10.811000,
                "C":12.010700, "N":14.006700, "O":15.999400, "F":18.998400, "Ne":20.179700,
                "Na":22.989700, "Mg":24.305000, "Al":26.981500, "Si":28.085500, "P":30.973800,
                "S":32.065000, "Cl":35.453000, "K":39.098300, "Ar":39.948000, "Ca":40.078000,
                "Sc":44.955900, "Ti":47.867000, "V":50.941500, "Cr":51.996100, "Mn":54.938000,
                "Fe":55.845000, "Ni":58.693400, "Co":58.933200, "Cu":63.546000, "Zn":65.390000,
                "Ga":69.723000, "Ge":72.640000, "As":74.921600, "Se":78.960000, "Br":79.904000,
                "Kr":83.800000, "Rb":85.467800, "Sr":87.620000, "Y":88.905900, "Zr":91.224000,
                "Nb":92.906400, "Mo":95.940000, "Tc":98.000000, "Ru":101.070000, "Rh":102.905500,
                "Pd":106.420000, "Ag":107.868200, "Cd":112.411000, "In":114.818000, "Sn":118.710000,
                "Sb":121.760000, "I":126.904500, "Te":127.600000, "Xe":131.293000, "Cs":132.905500,
                "Ba":137.327000, "La":138.905500, "Ce":140.116000, "Pr":140.907700, "Nd":144.240000,
                "Pm":145.000000, "Sm":150.360000, "Eu":151.964000, "Gd":157.250000, "Tb":158.925300,
                "Dy":162.500000, "Ho":164.930300, "Er":167.259000, "Tm":168.934200, "Yb":173.040000,
                "Lu":174.967000, "Hf":178.490000, "Ta":180.947900, "W":183.840000, "Re":186.207000,
                "Os":190.230000, "Ir":192.217000, "Pt":195.078000, "Au":196.966500, "Hg":200.590000,
                "Tl":204.383300, "Pb":207.200000, "Bi":208.980400, "Po":209.000000, "At":210.000000,
                "Rn":222.000000, "Fr":223.000000, "Ra":226.000000, "Ac":227.000000, "Pa":231.035900,
                "Th":232.038100, "Np":237.000000, "U":238.028900, "Am":243.000000, "Pu":244.000000,
                "Cm":247.000000, "Bk":247.000000, "Cf":251.000000, "Es":252.000000, "Fm":257.000000,
                "Md":258.000000, "No":259.000000, "Rf":261.000000, "Lr":262.000000, "Db":262.000000,
                "Bh":264.000000, "Sg":266.000000, "Mt":268.000000, "Rg":272.000000, "Hs":277.000000}

    # TO DO: allow customize mass
    massconvert = 0.000103642695727
    masses = {}
    for ele in ele_mass.keys():
        # Expects lines of format like: H 1.0 H_pseudo_name.ext
        masses[ele] = ele_mass[ele] * massconvert

    return positions, species, cell, masses


def dft_input_to_structure(dft_input: str):
    """
    Parses a qe input and returns the atoms in the file as a Structure object
    :param dft_input: QE Input file to parse
    :param cutoff: Cutoff radius for structure
    :return:
    """
    positions, species, cell, masses = parse_dft_input(dft_input)
    _, coded_species = struc.get_unique_species(species)
    return struc.Structure(positions=positions, species=coded_species,
                           cell=cell, mass_dict=masses, species_labels=species)

def edit_dft_input_positions(dft_input: str, structure):
    """
    Write the current configuration of the OTF structure to the
    qe input file
    """

    with open(dft_input, 'r') as f:
        lines = f.readlines()

    file_pos_index = None
    cell_index = None
    nat = None

    for i, line in enumerate(lines):
        if '&CELL' in line:
            cell_index = int(i + 1)
        if '&COORD' in line:
            file_pos_index = int(i + 1)
        if ('&END' in line and (file_pos_index is not None)):
            nat = i - file_pos_index
    #     if 'ATOMIC_SPECIES' in line:
    #         species_index = int(i + 1)

    assert file_pos_index is not None, 'Failed to find positions in input'
    assert cell_index is not None, 'Failed to find cell in input'
    assert nat is not None, 'Failed to find nat in input'

    # TODO Catch case where the punchout structure has more atoms than the
    # original structure

    for pos_index, line_index in enumerate(
            range(file_pos_index, file_pos_index + structure.nat)):
        pos_string = ' '.join([structure.species_labels[pos_index],
                               str(structure.positions[pos_index][
                                       0]),
                               str(structure.positions[pos_index][
                                       1]),
                               str(structure.positions[pos_index][
                                       2])])
        if line_index < len(lines):
            lines[line_index] = str(pos_string + '\n')
        else:
            lines.append(str(pos_string + '\n'))

    # # TODO current assumption: if there is a new structure, then the new
    # # structure has fewer atoms than the  previous one. If we are always
    # # 'editing' a version of the larger structure than this will be okay with
    # # the punchout method.
    # for line_index in range(file_pos_index + structure.nat,
    #                         file_pos_index + nat):
    #     lines[line_index] = ''

    lines[cell_index] = 'A '+' '.join([str(x) for x in structure.vec1]) + '\n'
    lines[cell_index + 1] = 'B '+' '.join([str(x) for x in structure.vec2]) \
                            + '\n'
    lines[cell_index + 2] = 'C '+' '.join([str(x) for x in structure.vec3]) \
                            + '\n'

    newfilename = dft_input+"_run"

    with open(newfilename, 'w') as f:
        for line in lines:
            f.write(line)

    return newfilename


def parse_dft_forces_and_energy(outfile: str):
    """
    Get forces from a pwscf file in eV/A

    :param outfile: str, Path to dft.output file
    :return: list[nparray] , List of forces acting on atoms
    """
    forces = []
    total_energy = np.nan

    startforce = -1
    with open(outfile, 'r') as outf:
        for line in outf:
            if line.find('FORCE_EVAL') != -1:
                total_energy = float(line.split()[8])

            if (startforce >= 2):
                if (line.find('SUM')!=-1):
                    startforce = -1
                else:
                    line = line.split()[3:]
                    forces.append(list(map(float, line)))
                    startforce += 1
            elif (startforce >=0):
                startforce += 1
            elif line.find('FORCES') != -1 and line.find('in') != -1:
                startforce = 0


    assert total_energy != np.nan, "dft parser failed to read " \
                                   "the file {}. Run failed.".format(outfile)

    # Convert from ry/au to ev/angstrom
    conversion_factor = 25.71104309541616*2.0

    forces = np.array(forces)*conversion_factor
    total_energy *= 27.2114

    return forces, total_energy



def parse_dft_forces(outfile: str):
    """
    Get forces from a pwscf file in eV/A

    :param outfile: str, Path to dft.output file
    :return: list[nparray] , List of forces acting on atoms
    """

    f, e = parse_dft_forces_and_energy(outfile)
    return f



import numpy as np


def update_positions(dt, noa, structure):
    dtdt = dt ** 2
    new_pos = np.zeros((noa, 3))

    for i, pre_pos in enumerate(structure.prev_positions):
        mass = structure.mass_dict[structure.species_labels[i]]
        pos = structure.positions[i]
        forces = structure.forces[i]

        new_pos[i] = 2 * pos - pre_pos + dtdt * forces / mass

    return new_pos


def calculate_temperature(new_pos, structure, dt, noa):
    # set velocity and temperature information
    velocities = (new_pos -
                  structure.prev_positions) / (2 * dt)

    KE = 0
    for i in range(len(structure.positions)):
        for j in range(3):
            KE += 0.5 * \
                structure.mass_dict[structure.species_labels[i]] * \
                velocities[i][j] * velocities[i][j]

    # see conversions.nb for derivation
    kb = 0.0000861733034

    # see p. 61 of "computer simulation of liquids"
    temperature = 2 * KE / ((3 * noa - 3) * kb)

    return KE, temperature, velocities
